import re

import allure
import pytest

from refinitiv.data.content.ownership._enums import Frequency
from refinitiv.data.content.ownership.consolidated import shareholders_history_report
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ownership.helpers import get_expected_column
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_dataframe_column_date_for_datetime_type,
)

EXPECTED_COLUMN_NAMES = [
    "calcdate",
    "investorid",
    "TR.InvestorFullName",
    "TR.InvParentType",
    "TR.InvestorType",
    "TR.InvestorRegion",
    "TR.InvInvestmentStyleCode",
    "TR.InvInvmtOrientation",
    "TR.OwnTrnverRating",
    "TR.OwnTurnover",
    "TR.TotalEquityAssets",
    "TR.SharesHeld",
    "TR.SharesHeldValue",
    "TR.PctOfSharesOutHeld",
    "TR.HoldingsDate",
    "TR.FilingType",
    "TR.SharesHeldChange",
    "TR.SharesHeldValChg",
    "TR.PrevSharesHeld",
    "TR.PrevSharesHeldValue",
    "TR.PrevPctOfSharesOutHeld",
    "TR.PrevHoldingsDate",
    "TR.PrevFilingType",
]

EXPECTED_COLUMN_TITLES = [
    "Calc Date",
    "Investor Id",
    "Investor Name",
    "Investor Parent Type",
    "Investor Type",
    "Investor Region",
    "Investment Style",
    "Orientation",
    "Turnover Rating",
    "Turnover %",
    "Total Equity Assets",
    "SharesHeld",
    "SharesHeld Value",
    "% SharesOutstanding",
    "Holdings Date",
    "Filing Type",
    "SharesHeld Change",
    "SharesHeld Value Change",
    "Previous SharesHeld",
    "Previous SharesHeld Value",
    "Previous % SharesOutstanding",
    "Previous Holdings Date",
    "Previous Filing Type",
]


@allure.suite("Content object - Ownership Consolidated Shareholders History Report")
@allure.feature("Content object - Ownership Consolidated Shareholders History Report")
@allure.severity(allure.severity_level.CRITICAL)
class TestConsolidatedShareholdersHistoryReport:
    @allure.title(
        "Create Consolidated Shareholders History Report definition object with valid required params and get data"
    )
    @pytest.mark.caseid("37556956")
    @pytest.mark.parametrize(
        "universe,frequency,use_field_names_in_headers",
        [("ONEX.CCP", "M", True), ("US02079K1079", Frequency.QUARTERLY, False)],
    )
    @pytest.mark.smoke
    def test_create_consolidated_shareholders_history_report_object_with_valid_fields_and_get_data(
        self, open_platform_session, universe, frequency, use_field_names_in_headers
    ):
        expected_column = get_expected_column(
            use_field_names_in_headers, EXPECTED_COLUMN_TITLES, EXPECTED_COLUMN_NAMES
        )
        response = shareholders_history_report.Definition(
            universe=universe,
            frequency=frequency,
            use_field_names_in_headers=use_field_names_in_headers,
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, expected_column)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title(
        "Create async Consolidated Shareholders History Report definition object with valid and invalid params"
    )
    @pytest.mark.caseid("37556957")
    @pytest.mark.parametrize(
        "universe,invalid_universe,frequency,start,end,limit",
        [
            ("ONEX.CCP", "invalid universe", "M", "20210725", "20211005", 4),
        ],
    )
    @pytest.mark.smoke
    async def test_async_consolidated_shareholders_history_report_object_with_valid_and_invalid_params(
        self,
        open_platform_session,
        universe,
        invalid_universe,
        frequency,
        start,
        end,
        limit,
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            shareholders_history_report.Definition(
                universe,
                frequency,
                start,
                end,
                use_field_names_in_headers=True,
                limit=limit,
            ),
            shareholders_history_report.Definition(
                invalid_universe,
                frequency,
                start,
                end,
                use_field_names_in_headers=True,
                limit=limit,
            ),
        )
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED_TWELVE,
            expected_error_message="Unable to resolve all requested identifiers. Requested items: ['invalid universe']",
        )
        check_response_dataframe_contains_columns_names(
            valid_response, EXPECTED_COLUMN_NAMES
        )

    @allure.title(
        "Create Consolidated Shareholders History Report definition object with invalid required params and get error"
    )
    @pytest.mark.caseid("37556958")
    @pytest.mark.parametrize(
        "universe,frequency",
        [("invalid", "INVALID FREQUENCY"), ("", 111)],
    )
    def test_create_consolidated_shareholders_history_report_object_with_invalid_fields_and_get_error(
        self, open_platform_session, universe, frequency
    ):
        with pytest.raises(
            AttributeError,
            match=re.escape(f"Value '{frequency}' must be in ['Q', 'M']"),
        ):
            shareholders_history_report.Definition(universe, frequency).get_data()



from ._american_monte_carlo_parameters import *
from ._amortization_item import *
from ._bid_ask_mid import *
from ._day_weight import *
from ._interpolation_weight import *
from ._pde_parameters import *
from ._basket_item import *
from ._bond_rounding_parameters import *
from ._numerical_method import *
from ._barrier_definition_element import *
from ._fx_point import *
from ._payout_scaling import *
from ._input_flow import *



import refinitiv.data.content.estimates as rde


def invalid_view_actuals_interim_estimates(universe):
    definition = rde.view_actuals.interim.Definition(
        universe=universe, package=rde.Package.PROFESSIONAL
    )
    return definition


def view_actuals_annual_estimates(universe, package=None):
    definition = rde.view_actuals.annual.Definition(
        universe=universe, package=package, use_field_names_in_headers=True
    ).get_data()
    return definition


def view_actuals_kpi_annual_estimates(universe, use_field_names_in_headers=False):
    definition = rde.view_actuals_kpi.annual.Definition(
        universe=universe, use_field_names_in_headers=use_field_names_in_headers
    ).get_data()
    return definition


def view_summary_annual_estimates(universe, package=None):
    definition = rde.view_summary.annual.Definition(
        universe=universe, package=package, use_field_names_in_headers=False
    ).get_data()
    return definition


def view_summary_historical_non_periodic_estimates(universe, package, extended_params):
    definition = rde.view_summary.historical_snapshots_non_periodic_measures.Definition(
        universe=universe, package=package, extended_params=extended_params
    ).get_data()
    return definition


def view_summary_historical_annual_estimates(universe, package, extended_params):
    definition = (
        rde.view_summary.historical_snapshots_periodic_measures_annual.Definition(
            universe=universe, package=package, extended_params=extended_params
        ).get_data()
    )
    return definition


def view_summary_recommendations_estimates(universe, package):
    definition = rde.view_summary.recommendations.Definition(
        universe=universe, package=package
    ).get_data()
    return definition


def view_summary_kpi_annual_estimates(universe):
    definition = rde.view_summary_kpi.annual.Definition(universe=universe)
    return definition



from refinitiv.data import _configure as configure


def test_get_config():
    # when
    result = configure.get_config()

    # then
    assert result is not None
    assert result is configure._config


def test_get_config_file():
    configure.reload()

    path = configure._get_filepath(None, None)

    assert path is None



from ..._curves._cross_currency_curves import MainConstituentAssetClass, RiskType
from ....._tools import EnumArgsParser, make_parse_enum, make_convert_to_enum


main_constituent_asset_class_arg_parser = EnumArgsParser(
    parse=make_parse_enum(MainConstituentAssetClass),
    parse_to_enum=make_convert_to_enum(MainConstituentAssetClass),
)

risk_type_arg_parser = EnumArgsParser(
    parse=make_parse_enum(RiskType),
    parse_to_enum=make_convert_to_enum(RiskType),
)



from typing import Optional, TYPE_CHECKING

from ._fx_forward_turn_fields import FxForwardTurnFields
from ...._object_definition import ObjectDefinition

if TYPE_CHECKING:
    from ......_types import OptStr


class OverrideFxForwardTurn(ObjectDefinition):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    start_date : str, optional

    end_date : str, optional

    fields : FxForwardTurnFields, optional

    date : str, optional

    turn_tag : str, optional

    """

    def __init__(
        self,
        start_date: "OptStr" = None,
        end_date: "OptStr" = None,
        fields: Optional[FxForwardTurnFields] = None,
        date: "OptStr" = None,
        turn_tag: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.start_date = start_date
        self.end_date = end_date
        self.fields = fields
        self.date = date
        self.turn_tag = turn_tag

    @property
    def fields(self):
        """
        :return: object FxForwardTurnFields
        """
        return self._get_object_parameter(FxForwardTurnFields, "fields")

    @fields.setter
    def fields(self, value):
        self._set_object_parameter(FxForwardTurnFields, "fields", value)

    @property
    def date(self):
        """
        :return: str
        """
        return self._get_parameter("date")

    @date.setter
    def date(self, value):
        self._set_parameter("date", value)

    @property
    def end_date(self):
        """
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def start_date(self):
        """
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        self._set_parameter("startDate", value)

    @property
    def turn_tag(self):
        """
        :return: str
        """
        return self._get_parameter("turnTag")

    @turn_tag.setter
    def turn_tag(self, value):
        self._set_parameter("turnTag", value)



import datetime
import time
import logging
from collections import namedtuple
from typing import Callable, Optional, Dict

from refinitiv.data._core.log_reporter import LogReporter
from refinitiv.data._core.session import Session
from refinitiv.data._errors import SessionError

try:
    from ema import (
        OmmConsumer,
        OmmConsumerConfig,
        AppClient,
        CustomLoggerClient,
        LoggerSeverity,
        MapEntry,
        Msg,
        OmmConsumerEvent,
        RefreshMsg,
        LoginRefresh,
        EmaOmmInvalidHandleException,
        EmaOmmInvalidUsageException,
    )  # noqa

    EMA_INSTALLED = True
except ImportError as e:
    EMA_INSTALLED = False
    _exc_info = e
else:
    from .conversion import json_marketprice_msg_to_ema
    from .ema import (
        ema_login_message,
        create_programmatic_cfg,
        name_type_map,
        generate_login_msg,
    )


if EMA_INSTALLED:

    class EmaPythonLogger(CustomLoggerClient):
        def __init__(self):
            super().__init__(EmaPythonLogger.log)

        severity_logging_map: dict = {
            LoggerSeverity.NoLogMsg: 1000,  # higher than critical to avoid logging
            LoggerSeverity.Error: logging.ERROR,
            LoggerSeverity.Warning: logging.WARNING,
            LoggerSeverity.Success: logging.INFO,
            LoggerSeverity.Verbose: logging.DEBUG,
        }

        @staticmethod
        def log(callback_client_name: str, severity: LoggerSeverity, message: str):
            logging.log(
                EmaPythonLogger.severity_logging_map[severity],
                f"PythonLogger - {callback_client_name}: {message}",
            )


handle_tuple = namedtuple("handle_tuple", ["handle", "msg"])


class RwfSocketClient(LogReporter):
    """Emulating WebsocketApp + WSAPI"""

    def __init__(
        self,
        host: str,
        port: int,
        on_open: Callable,
        on_message: Callable,
        on_close: Callable,
        session: "Session",
        field_dict_path: Optional[str] = None,
        enumtype_path: Optional[str] = None,
        python_logger: bool = True,
    ):
        LogReporter.__init__(self, logger=session.logger())
        if not EMA_INSTALLED:
            self.debug(f"EMA not installed, message: {_exc_info}")
            raise ImportError("You need to install refinitiv-ema to use RwfSocketApp")

        self.host = host
        self.port = port

        self.field_dict_path = field_dict_path
        self.enumtype_path = enumtype_path
        self.on_open = on_open
        self.on_message = on_message
        self.on_close = on_close
        self.id = None
        self.handles: Dict[int, handle_tuple] = {}  # stream_id: (handle, msg)
        self.consumer: Optional[OmmConsumer] = None
        self.keep_running = False
        self._reissue_handle = None
        self._reissue_timestamp = 0

        self.client = AppClient(
            on_refresh_msg=self.__msg_cb,
            on_update_msg=self.__msg_cb,
            on_status_msg=self.__msg_cb,
        )

        self.login_client = AppClient(
            on_refresh_msg=self.__refresh_msg_cb,
            on_update_msg=self.__nothing_callback,
            on_status_msg=self.__msg_cb,
        )
        if python_logger:
            self.ema_logger = EmaPythonLogger()
        else:
            self.ema_logger = None  # Will use internal OmmLoggerClient

    def __refresh_msg_cb(self, msg: "RefreshMsg", event: "OmmConsumerEvent"):
        self._reissue_handle = event.get_handle()
        tmp_refresh = LoginRefresh(msg)
        if tmp_refresh.has_authentication_tt_reissue():
            self._reissue_timestamp = tmp_refresh.get_authentication_tt_reissue()
        else:
            self._reissue_timestamp = 0

    def __msg_cb(self, msg: "Msg", event: "OmmConsumerEvent"):
        msg_dict = msg.to_dict()
        # We swap ID of the stream with the one from closure because EMA ignores
        # stream id parameter
        msg_dict["ID"] = event.get_closure()
        self.on_message([msg_dict])

    def __nothing_callback(self, msg: "Msg", _: "OmmConsumerEvent"):
        self.debug(f"Ignored message {msg.to_dict()}")

    def close(self, **_):
        self.keep_running = False
        self.on_close(self, -100, "Just Closed")
        # I don't think this is the best way, but at least it works
        if self.client:
            self.client.on_refresh_msg = None
            self.client.on_update_msg = None
            self.client.on_status_msg = None
        if self.login_client:
            self.login_client.on_refresh_msg = None
            self.login_client.on_update_msg = None
            self.login_client.on_status_msg = None

    def run_forever(self):
        self.on_open(self)
        self.keep_running = True

        try:
            while self.keep_running:
                time.sleep(0.2)

        except (Exception, KeyboardInterrupt, SystemExit) as e:
            # self._callback(self.on_error, e)
            if isinstance(e, SystemExit):
                # propagate SystemExit further
                raise
            # teardown()
            return not isinstance(e, KeyboardInterrupt)

    def _init_consumer(self, login_msg: dict):
        admin_msg = ema_login_message(**generate_login_msg(login_msg))
        pgcfg = create_programmatic_cfg(
            field_dict_path=self.field_dict_path,
            enumtype_path=self.enumtype_path,
            host=self.host,
            port=self.port,
        )
        config = OmmConsumerConfig().config(pgcfg).add_admin_msg(admin_msg)
        try:
            self.consumer = OmmConsumer(
                config,
                self.login_client,
                self.ema_logger,
            )
        except EmaOmmInvalidUsageException as e:
            self.close()
            raise SessionError(
                -1,
                "Error establishing connection to RSSL endpoint. "
                "Check the logs for more details",
            ) from e

    def _handle_reissue(self, msg):
        reissue_success = False
        if self._reissue_handle is not None and self._reissue_timestamp >= time.time():
            try:
                admin_msg = ema_login_message(**generate_login_msg(msg))
                self.consumer.reissue(admin_msg, self._reissue_handle)
                reissue_success = True
                self.info(
                    f"Sent reissue with handle {self._reissue_handle}"
                    f" and timestamp {self._reissue_timestamp}"
                )
            except EmaOmmInvalidHandleException:
                self.warning(
                    f"Reissue failed: reissue was valid till "
                    f"{datetime.datetime.fromtimestamp(self._reissue_timestamp)}.\n"
                    f"This is likely because connection was broken "
                    f"for more than 10 minutes."
                )
        return reissue_success

    def _handle_consumer(self, msg):
        self.info("Creating consumer")
        self._init_consumer(msg)
        if len(self.handles) > 0:
            self.info("Re-registering all streams")
            for _, msg in self.handles.copy().values():
                self._send_message(msg)

    def _send_message(self, msg: dict):
        msg_id = msg.pop("ID")
        handle = self.consumer.register_client(
            json_marketprice_msg_to_ema(msg, self.consumer.field_id_map),
            self.client,
            msg_id,
        )
        self.handles[msg_id] = handle_tuple(handle, msg)

    def send(self, msg: dict):
        msg_type = msg.get("Type", "Request")
        msg_domain = msg.get("Domain", "MarketPrice")

        if msg_type == "Pong":
            # do nothing
            pass

        elif msg_type == "Request":  # open
            if msg_domain == "MarketPrice":
                self._send_message(msg)
                # logger.debug(f"new client registered for {msg['Key']['Name']}")
            elif msg_domain == "Login":
                if not self._handle_reissue(msg):
                    self._handle_consumer(msg)
            else:
                raise ValueError(f"Unknown domain of Request message: {msg_domain}")

        elif msg_type == "Close":
            if msg_domain == "Login":
                self.close()  # TODO: is this right?
            else:
                self.debug(f"Closing stream {msg['ID']}")
                self.consumer.unregister(self.handles[msg["ID"]].handle)
                del self.handles[msg["ID"]]
        else:
            raise ValueError(f"Unknown message type: {msg_type}")



from datetime import timedelta, datetime

import allure
import pytest

from refinitiv.data.content.historical_pricing import events, Adjustments, EventTypes
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.historical_pricing.events.conftest import (
    check_response_data_fields,
    check_response_data_event_type,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_non_empty_response_data,
    check_response_status,
    assert_error,
    check_response_data_start_end_date,
    is_universe_empty,
    check_response_is_success,
    check_response_data_universe,
    is_expected_param_in_request_url,
    check_extended_params_were_sent_in_request,
)


@allure.suite("Historical Pricing Events")
@allure.feature("Historical Pricing Events")
@allure.severity(allure.severity_level.CRITICAL)
class TestHistoricalPricingEvents:
    @allure.title("Pricing event definition - get data")
    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "universe, fields, count",
        [[["VOD.L", "GOOG.O"], ["BID", "BIDSIZE", "ASK"], 42]],
    )
    @pytest.mark.caseid("С37902412")
    def test_get_data(self, open_desktop_session, universe, fields, count):
        response = events.Definition(universe, fields=fields, count=count).get_data()
        check_response_is_success(response)
        check_response_data_universe(response, universe)
        check_response_data_fields(response, fields)
        assert len(response.data.raw[0]["data"]) == count, (
            f"Pricing dataframe size is not as expected: "
            f"{response.data.raw[0]['data']} != {count}"
        )

    @allure.title("Pricing event definition - get data with interval")
    @pytest.mark.parametrize(
        "universe, fields, start_date, end_date, event_type",
        [
            (
                "LSEG.L",
                ["EVENT_TYPE", "BID", "ASK"],
                timedelta(-1),
                timedelta(0),
                "trade",
            ),
            ("VOD.L", None, timedelta(hours=-24), datetime.now(), EventTypes.QUOTE),
            ("EUR=", None, "2022-12-18T10:00:00", "2022-12-19T16:00:00", "quote"),
        ],
        ids=["timedelta", "datetime", "string"],
    )
    @pytest.mark.caseid("C37902415")
    def test_get_data_with_interval(
        self,
        open_desktop_session,
        universe,
        fields,
        start_date,
        end_date,
        event_type,
        request,
    ):
        response = events.Definition(
            universe,
            fields=fields,
            start=start_date,
            end=end_date,
            eventTypes=event_type,
            count=13000,
        ).get_data()
        is_expected_param_in_request_url(response, ["start", "end"])
        check_response_is_success(response)
        check_response_data_start_end_date(response, start_date, end_date, request)
        check_response_data_event_type(response, event_type)

    @allure.title(
        "Pricing event definition - get data with mix of valid and invalid universe"
    )
    @pytest.mark.parametrize(
        "universe",
        [["IBM", "WRONG_UNIVERSE"]],
    )
    @pytest.mark.caseid("C38941492")
    def test_get_data_with_mix_of_valid_and_invalid_universe(
        self,
        open_desktop_session,
        universe,
    ):
        response = events.Definition(universe).get_data()
        check_response_is_success(response)
        assert not is_universe_empty(response.data.df, universe[0])
        assert is_universe_empty(response.data.df, universe[1])

    @allure.title("Pricing event definition - get data without universe")
    @pytest.mark.caseid("C37913678")
    def test_get_data_without_universe(self):
        with pytest.raises(
            TypeError,
        ) as error:
            events.Definition()

        assert_error(error, "universe")

    @allure.title("Pricing event definition - get data with incorrect parameters")
    @pytest.mark.parametrize(
        "universe, fields, start, end, event_type",
        [["123", ["QWE"], "2021-12-01T00:00:00", "2020-12-07T00:10:00", "asd"]],
    )
    @pytest.mark.caseid("C37902416")
    def test_get_data_with_incorrect_parameters(
        self, open_desktop_session, universe, fields, start, end, event_type
    ):
        with pytest.raises(RDError) as error:
            events.Definition("IBM", fields=fields).get_data()
        assert "The universe does not support the following fields" in str(error.value)

        with pytest.raises(RDError) as error:
            events.Definition("IBM", start=start, end=end).get_data()
        assert "Request Validation Error" in str(error.value)

    @allure.title(
        "Create Historical Pricing Events get data async with valid and invalid universe"
    )
    @pytest.mark.parametrize("universe,invalid_universe", [("VOD.L", "INVALID")])
    @pytest.mark.caseid("C38941498")
    async def test_get_data_async(
        self, open_desktop_session_async, universe, invalid_universe
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            events.Definition(universe), events.Definition(invalid_universe)
        )
        check_response_is_success(valid_response)
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="TS.Intraday.UserRequestError.90001",
            expected_error_message="INVALID - The universe is not found",
        )

    @allure.title("Pricing event definition - get partial data with extended params")
    @pytest.mark.parametrize(
        "universe,fields,adjustments,extended_params",
        [
            (
                "GBP=",
                ["BID", "ASK"],
                None,
                {"count": "50", "adjustments": "exchangeCorrection,manualCorrection"},
            ),
            (
                "GBP=",
                "ASK",
                [
                    Adjustments.EXCHANGE_CORRECTION,
                    Adjustments.MANUAL_CORRECTION,
                    Adjustments.QUALIFIERS,
                ],
                None,
            ),
        ],
    )
    @pytest.mark.caseid("C37902415")
    def test_get_data_with_extended_params_and_partial_data(
        self, open_desktop_session, universe, fields, adjustments, extended_params
    ):
        response = events.Definition(
            universe=universe,
            fields=fields,
            adjustments=adjustments,
            start=timedelta(hours=-1),
            end=timedelta(0),
            extended_params=extended_params,
        ).get_data()
        check_response_is_success(response)
        check_non_empty_response_data(response)
        check_response_status(
            response=response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="TS.Intraday.Warning.95004",
            expected_error_message=f"{universe} - Trades interleaving with corrections is currently not supported. Corrections will not be returned.",
        )
        check_extended_params_were_sent_in_request(response, extended_params)



from refinitiv.data.content.ipa.curves._cross_currency_curves import definitions
from refinitiv.data.content.ipa.curves._cross_currency_curves.definitions import manage
from refinitiv.data.delivery._data._response import Response
from tests.unit.conftest import StubResponse, StubSession


def test_update_function():
    # given
    response = StubResponse(
        {
            "curveDefinition": {
                "baseCurrency": "EUR",
                "baseIndexName": "ESTR",
                "name": "rename curve",
                "quotedCurrency": "USD",
                "quotedIndexName": "SOFR",
                "source": "SourceName",
                "isNonDeliverable": False,
                "mainConstituentAssetClass": "FxForward",
                "riskType": "CrossCurrency",
                "isFallbackForFxCurveDefinition": False,
                "firstHistoricalAvailabilityDate": "2021-01-01",
                "id": "7bdb00f3-0a48-40be-ace2-6d3cfd0e8e59",
            },
            "curveInfo": {
                "creationDateTime": "2022-12-08T03:19:00.011739Z",
                "creationUserId": "GESG1-111923",
                "updateDateTime": "2022-12-08T03:31:56.915952Z",
                "updateUserId": "GESG1-111923",
            },
            "segments": [
                {
                    "constituents": {
                        "fxForwards": [
                            {
                                "instrumentDefinition": {
                                    "instrumentCode": "EUR1M=",
                                    "tenor": "1M",
                                    "isNonDeliverable": False,
                                    "quotationMode": "Outright",
                                }
                            }
                        ]
                    },
                    "startDate": "2021-01-01",
                }
            ],
        }
    )
    session = StubSession(is_open=True, response=response)
    response = manage.update(
        session=session,
        curve_definition=definitions.CrossCurrencyCurveUpdateDefinition(
            id="7bdb00f3-0a48-40be-ace2-6d3cfd0e8e59",
            source="SourceName",
            name="rename curve",
            base_currency="EUR",
            base_index_name="ESTR",
            quoted_currency="USD",
            quoted_index_name="SOFR",
            is_non_deliverable=False,
        ),
        segments=[
            definitions.CrossCurrencyInstrumentsSegment(
                start_date="2021-01-01",
                constituents=definitions.CrossCurrencyConstituentsDescription(
                    fx_forwards=[
                        definitions.FxForwardInstrumentDescription(
                            instrument_definition=definitions.FxForwardInstrumentDefinition(
                                instrument_code="EUR1M=",
                                tenor="1M",
                                is_non_deliverable=False,
                                quotation_mode=definitions.QuotationMode.OUTRIGHT,
                            )
                        )
                    ]
                ),
            )
        ],
    )

    # then
    assert isinstance(response, Response)



# coding: utf8

from typing import Optional, List

from .._base import UnderlyingDefinition
from ..._instrument_definition import InstrumentDefinition
from .._enums import (
    BuySell,
    CallPut,
    ExerciseStyle,
    UnderlyingType,
    SettlementType,
)
from . import (
    FxDualCurrencyDefinition,
    FxAverageInfo,
    FxBarrierDefinition,
    FxBinaryDefinition,
    FxDoubleBarrierDefinition,
    FxDoubleBinaryDefinition,
    FxForwardStart,
    FxUnderlyingDefinition,
)
from .._models import InputFlow


class FxDefinition(InstrumentDefinition):
    """
    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.it can be used to link output
        results to the instrument definition. only alphabetic, numeric and '- _.#=@'
        characters are supported. optional.
    start_date : str, optional
        Start date of the option
    end_date : str, optional
        The maturity or expiry date of the instrument. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z').
        optional. mandatory for otc eti options and fx options(if tenor is not defined).
        if instrumentcode of listed eti option is defined, the value comes from the
        instrument reference data.
    tenor : str, optional
        The code indicating the period between startdate and enddate of the instrument
        (e.g. '6m', '1y')
    notional_ccy : str, optional
        The currency of the instrument's notional amount. the value is expressed in iso
        4217 alphabetical format (e.g. 'usd'). if the option is a eurgbp call option,
        notionalccy can be expressed in eur or gbp mandatory for fx options.
    notional_amount : float, optional
        The notional amount of the instrument. if the option is a eurgbp call option,
        amount of eur or gbp of the contract
    asian_definition : FxOptionAverageInfo, optional

    barrier_definition : FxOptionBarrierDefinition, optional

    binary_definition : FxOptionBinaryDefinition, optional

    buy_sell : BuySell, optional
        The indicator of the deal side. the possible values are:
        - buy: buying the option,
        - sell: selling/writing the option. the output amounts calculated with taking
          buysell into consideration are returned with a reversed sign when the value
          'sell' is used. optional. the default value is 'buy'.
    call_put : CallPut, optional
        The indicator if the option is a call or a put. the possible values are:
        - call: the right to buy the underlying asset,
        - put: the right to sell the underlying asset. optional. if instrumentcode of
          listed eti option is defined, the value comes from the instrument reference
          data.the default value is 'call' for otc eti options and fx options.
    double_barrier_definition : FxOptionDoubleBarrierDefinition, optional

    double_binary_definition : FxOptionDoubleBinaryDefinition, optional

    dual_currency_definition : FxDualCurrencyDefinition, optional

    exercise_style : ExerciseStyle, optional
        The option style based on its exercise restrictions. the possible values are:
        - amer: the owner has the right to exercise on any date before the option
          expires,
        - euro: the owner has the right to exercise only on enddate,
        - berm: the owner has the right to exercise on any of several specified dates
          before the option expires. all exercise styles may not apply to certain option
          types. optional. if instrumentcode of listed eti option is defined, the value
          comes from the instrument reference data. the default value is 'euro' for otc
          eti options and fx options.
    forward_start_definition : FxOptionForwardStart, optional

    payments : InputFlow, optional
        An array of payments
    settlement_type : SettlementType, optional
        The settlement method for options when exercised. the possible values are:
        - physical(asset): delivering the underlying asset.
        - cash: paying out in cash.
    underlying_definition : FxUnderlyingDefinition, optional

    underlying_type : UnderlyingType, optional
        The type of the option based on the underlying asset. the possible values are:
        - eti: eti(exchanged traded instruments) options,
        - fx: fx options. mandatory. no default value applies.
    delivery_date : str, optional
        The date when the underlylng asset is delivered. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g. '2021-01-01t00:00:00z').
    settlement_ccy : str, optional
        The currency of the instrument's settlement. the value is expressed in iso 4217
        alphabetical format (e.g. 'usd'). if the option is a eurgbp call option,
        settlementccy can be expressed in eur or gbp
    strike : float, optional
        The set price at which the owner of the option can buy or sell the underlying
        asset. the value is expressed according to the market convention linked to the
        underlying asset. optional. mandatory for otc eti options and fx options. if
        instrumentcode of listed eti option is defined, the value comes from the
        instrument reference data.
    """

    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        notional_ccy: Optional[str] = None,
        notional_amount: Optional[float] = None,
        asian_definition: Optional[FxAverageInfo] = None,
        barrier_definition: Optional[FxBarrierDefinition] = None,
        binary_definition: Optional[FxBinaryDefinition] = None,
        buy_sell: Optional[BuySell] = None,
        call_put: Optional[CallPut] = None,
        double_barrier_definition: Optional[FxDoubleBarrierDefinition] = None,
        double_binary_definition: Optional[FxDoubleBinaryDefinition] = None,
        dual_currency_definition: Optional[FxDualCurrencyDefinition] = None,
        exercise_style: Optional[ExerciseStyle] = None,
        forward_start_definition: Optional[FxForwardStart] = None,
        payments: Optional[List[InputFlow]] = None,
        settlement_type: Optional[SettlementType] = None,
        underlying_definition: Optional[FxUnderlyingDefinition] = None,
        underlying_type: Optional[UnderlyingType] = None,
        delivery_date: Optional[str] = None,
        settlement_ccy: Optional[str] = None,
        strike: Optional[float] = None,
        **kwargs,
    ) -> None:
        super().__init__(instrument_tag, **kwargs)
        self.instrument_tag = instrument_tag
        self.start_date = start_date
        self.end_date = end_date
        self.tenor = tenor
        self.notional_ccy = notional_ccy
        self.notional_amount = notional_amount
        self.asian_definition = asian_definition
        self.barrier_definition = barrier_definition
        self.binary_definition = binary_definition
        self.buy_sell = buy_sell
        self.call_put = call_put
        self.double_barrier_definition = double_barrier_definition
        self.double_binary_definition = double_binary_definition
        self.dual_currency_definition = dual_currency_definition
        self.exercise_style = exercise_style
        self.forward_start_definition = forward_start_definition
        self.payments = payments
        self.settlement_type = settlement_type
        self.underlying_definition = underlying_definition
        self.underlying_type = underlying_type
        self.delivery_date = delivery_date
        self.settlement_ccy = settlement_ccy
        self.strike = strike

    @property
    def asian_definition(self):
        """
        :return: object FxOptionAverageInfo
        """
        return self._get_object_parameter(FxAverageInfo, "asianDefinition")

    @asian_definition.setter
    def asian_definition(self, value):
        self._set_object_parameter(FxAverageInfo, "asianDefinition", value)

    @property
    def barrier_definition(self):
        """
        :return: object FxOptionBarrierDefinition
        """
        return self._get_object_parameter(FxBarrierDefinition, "barrierDefinition")

    @barrier_definition.setter
    def barrier_definition(self, value):
        self._set_object_parameter(FxBarrierDefinition, "barrierDefinition", value)

    @property
    def binary_definition(self):
        """
        :return: object FxOptionBinaryDefinition
        """
        return self._get_object_parameter(FxBinaryDefinition, "binaryDefinition")

    @binary_definition.setter
    def binary_definition(self, value):
        self._set_object_parameter(FxBinaryDefinition, "binaryDefinition", value)

    @property
    def buy_sell(self):
        """
        The indicator of the deal side. the possible values are:
        - buy: buying the option,
        - sell: selling/writing the option. the output amounts calculated with taking
          buysell into consideration are returned with a reversed sign when the value
          'sell' is used. optional. the default value is 'buy'.
        :return: enum BuySell
        """
        return self._get_enum_parameter(BuySell, "buySell")

    @buy_sell.setter
    def buy_sell(self, value):
        self._set_enum_parameter(BuySell, "buySell", value)

    @property
    def call_put(self):
        """
        The indicator if the option is a call or a put. the possible values are:
        - call: the right to buy the underlying asset,
        - put: the right to sell the underlying asset. optional. if instrumentcode of
          listed eti option is defined, the value comes from the instrument reference
          data.the default value is 'call' for otc eti options and fx options.
        :return: enum CallPut
        """
        return self._get_enum_parameter(CallPut, "callPut")

    @call_put.setter
    def call_put(self, value):
        self._set_enum_parameter(CallPut, "callPut", value)

    @property
    def double_barrier_definition(self):
        """
        :return: object FxOptionDoubleBarrierDefinition
        """
        return self._get_object_parameter(
            FxDoubleBarrierDefinition, "doubleBarrierDefinition"
        )

    @double_barrier_definition.setter
    def double_barrier_definition(self, value):
        self._set_object_parameter(
            FxDoubleBarrierDefinition, "doubleBarrierDefinition", value
        )

    @property
    def double_binary_definition(self):
        """
        :return: object FxOptionDoubleBinaryDefinition
        """
        return self._get_object_parameter(
            FxDoubleBinaryDefinition, "doubleBinaryDefinition"
        )

    @double_binary_definition.setter
    def double_binary_definition(self, value):
        self._set_object_parameter(
            FxDoubleBinaryDefinition, "doubleBinaryDefinition", value
        )

    @property
    def dual_currency_definition(self):
        """
        :return: object FxDualCurrencyDefinition
        """
        return self._get_object_parameter(
            FxDualCurrencyDefinition, "dualCurrencyDefinition"
        )

    @dual_currency_definition.setter
    def dual_currency_definition(self, value):
        self._set_object_parameter(
            FxDualCurrencyDefinition, "dualCurrencyDefinition", value
        )

    @property
    def exercise_style(self):
        """
        The option style based on its exercise restrictions. the possible values are:
        - amer: the owner has the right to exercise on any date before the option
          expires,
        - euro: the owner has the right to exercise only on enddate,
        - berm: the owner has the right to exercise on any of several specified dates
          before the option expires. all exercise styles may not apply to certain option
          types. optional. if instrumentcode of listed eti option is defined, the value
          comes from the instrument reference data. the default value is 'euro' for otc
          eti options and fx options.
        :return: enum ExerciseStyle
        """
        return self._get_enum_parameter(ExerciseStyle, "exerciseStyle")

    @exercise_style.setter
    def exercise_style(self, value):
        self._set_enum_parameter(ExerciseStyle, "exerciseStyle", value)

    @property
    def forward_start_definition(self):
        """
        :return: object FxOptionForwardStart
        """
        return self._get_object_parameter(FxForwardStart, "forwardStartDefinition")

    @forward_start_definition.setter
    def forward_start_definition(self, value):
        self._set_object_parameter(FxForwardStart, "forwardStartDefinition", value)

    @property
    def payments(self):
        """
        An array of payments
        :return: list InputFlow
        """
        return self._get_list_parameter(InputFlow, "payments")

    @payments.setter
    def payments(self, value):
        self._set_list_parameter(InputFlow, "payments", value)

    @property
    def settlement_type(self):
        """
        The settlement method for options when exercised. the possible values are:
        - physical(asset): delivering the underlying asset.
        - cash: paying out in cash.
        :return: enum SettlementType
        """
        return self._get_enum_parameter(SettlementType, "settlementType")

    @settlement_type.setter
    def settlement_type(self, value):
        self._set_enum_parameter(SettlementType, "settlementType", value)

    @property
    def underlying_definition(self):
        """
        :return: object FxUnderlyingDefinition
        """
        return self._get_object_parameter(UnderlyingDefinition, "underlyingDefinition")

    @underlying_definition.setter
    def underlying_definition(self, value):
        self._set_object_parameter(UnderlyingDefinition, "underlyingDefinition", value)

    @property
    def underlying_type(self):
        """
        The type of the option based on the underlying asset. the possible values are:
        - eti: eti(exchanged traded instruments) options,
        - fx: fx options. mandatory. no default value applies.
        :return: enum UnderlyingType
        """
        return self._get_enum_parameter(UnderlyingType, "underlyingType")

    @underlying_type.setter
    def underlying_type(self, value):
        self._set_enum_parameter(UnderlyingType, "underlyingType", value)

    @property
    def delivery_date(self):
        """
        The date when the underlylng asset is delivered. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g. '2021-01-01t00:00:00z').
        :return: str
        """
        return self._get_parameter("deliveryDate")

    @delivery_date.setter
    def delivery_date(self, value):
        self._set_parameter("deliveryDate", value)

    @property
    def end_date(self):
        """
        The maturity or expiry date of the instrument. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z').
        optional. mandatory for otc eti options and fx options(if tenor is not defined).
        if instrumentcode of listed eti option is defined, the value comes from the
        instrument reference data.
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def instrument_tag(self):
        """
        User defined string to identify the instrument.it can be used to link output
        results to the instrument definition. only alphabetic, numeric and '- _.#=@'
        characters are supported. optional.
        :return: str
        """
        return self._get_parameter("instrumentTag")

    @instrument_tag.setter
    def instrument_tag(self, value):
        self._set_parameter("instrumentTag", value)

    @property
    def notional_amount(self):
        """
        The notional amount of the instrument. if the option is a eurgbp call option,
        amount of eur or gbp of the contract
        :return: float
        """
        return self._get_parameter("notionalAmount")

    @notional_amount.setter
    def notional_amount(self, value):
        self._set_parameter("notionalAmount", value)

    @property
    def notional_ccy(self):
        """
        The currency of the instrument's notional amount. the value is expressed in iso
        4217 alphabetical format (e.g. 'usd'). if the option is a eurgbp call option,
        notionalccy can be expressed in eur or gbp mandatory for fx options.
        :return: str
        """
        return self._get_parameter("notionalCcy")

    @notional_ccy.setter
    def notional_ccy(self, value):
        self._set_parameter("notionalCcy", value)

    @property
    def settlement_ccy(self):
        """
        The currency of the instrument's settlement. the value is expressed in iso 4217
        alphabetical format (e.g. 'usd'). if the option is a eurgbp call option,
        settlementccy can be expressed in eur or gbp
        :return: str
        """
        return self._get_parameter("settlementCcy")

    @settlement_ccy.setter
    def settlement_ccy(self, value):
        self._set_parameter("settlementCcy", value)

    @property
    def start_date(self):
        """
        Start date of the option
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        self._set_parameter("startDate", value)

    @property
    def strike(self):
        """
        The set price at which the owner of the option can buy or sell the underlying
        asset. the value is expressed according to the market convention linked to the
        underlying asset. optional. mandatory for otc eti options and fx options. if
        instrumentcode of listed eti option is defined, the value comes from the
        instrument reference data.
        :return: float
        """
        return self._get_parameter("strike")

    @strike.setter
    def strike(self, value):
        self._set_parameter("strike", value)

    @property
    def tenor(self):
        """
        The code indicating the period between startdate and enddate of the instrument
        (e.g. '6m', '1y')
        :return: str
        """
        return self._get_parameter("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_parameter("tenor", value)



from datetime import datetime, timedelta

import allure
import pytest

import refinitiv.data.eikon as ek
from refinitiv.data.errors import RDError
from tests.integration.eikon.time_series.conftest import (
    check_dataframe_contains_ohlc_columns_and_rics,
)


@allure.suite("Eikon Legacy module - Timeseries")
@allure.feature("Eikon Legacy module - Timeseries")
@allure.severity(allure.severity_level.NORMAL)
class TestTimeSeries:
    @allure.title("Get timeseries with single ric")
    @pytest.mark.caseid("38427366")
    @pytest.mark.parametrize("ric", ["VOD.L"])
    @pytest.mark.smoke
    def test_get_timeseries_with_single_ric(self, setup_app_key, ric):
        df = ek.get_timeseries(ric)
        assert not df.empty, "Empty dataframe received"
        assert df.columns.name == ric
        assert df.columns.array == ["HIGH", "LOW", "OPEN", "CLOSE", "COUNT", "VOLUME"]

    @allure.title("Get timeseries with optional parameters")
    @pytest.mark.caseid("38427367")
    @pytest.mark.parametrize(
        "ric,start_date,end_date,interval",
        [
            (
                ["VOD.L", "LSEG.L"],
                timedelta(-10),
                timedelta(0),
                "daily",
            ),
            (
                "VOD.L",
                timedelta(-180),
                datetime.now(),
                "monthly",
            ),
        ],
    )
    def test_get_timeseries_with_optional_parameters(
        self, setup_app_key, ric, start_date, end_date, interval
    ):
        df = ek.get_timeseries(
            ric, start_date=start_date, end_date=end_date, interval=interval
        )
        assert not df.empty, "Empty dataframe received"
        check_dataframe_contains_ohlc_columns_and_rics(df, ric)

    @allure.title("Get timeseries with invalid ric")
    @pytest.mark.caseid("38427368")
    @pytest.mark.parametrize("ric", ["invalid_ric"])
    def test_get_timeseries_with_invalid_ric(self, setup_app_key, ric):
        with pytest.raises(RDError, match="Error code Error | INVALID: Invalid RIC |"):
            ek.get_timeseries(ric)



from enum import Enum

import pytest

from refinitiv.data._content_type import ContentType
from refinitiv.data._core.session import SessionType
from refinitiv.data.content.fundamental_and_reference import RowHeaders
from refinitiv.data.content.fundamental_and_reference._definition import (
    row_headers_arg_parser,
)


def create_definition_and_session(platform, session_type):
    config = StubConfig(
        {
            "apis.data.datagrid": {
                "url": "/data/datagrid/beta1",
                "underlying-platform": platform,
                "endpoints.standard": "/",
            },
            "raise_exception_on_error": True,
        }
    )
    response = StubResponse(content_data={"responses": [fund_and_ref_response]})
    session = StubSession(response=response, config=config, is_open=True)
    session.type = session_type
    definition = fundamental_and_reference.Definition([], [], row_headers="date")
    return definition, session


class EnumTest(Enum):
    TEST = "test"
    DATE = "date"


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        (None, []),
        ([], []),
        ("date", [RowHeaders.DATE]),
        (["date"], [RowHeaders.DATE]),
        (RowHeaders.DATE.value, [RowHeaders.DATE]),
        (RowHeaders.DATE, [RowHeaders.DATE]),
    ],
)
def test_parse_row_headers_correct(input_value, expected_value):
    # when
    testing_value = row_headers_arg_parser.get_list(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_value",
    [
        1,
        {},
        EnumTest,
        EnumTest.TEST,
        EnumTest.TEST.value,
        EnumTest.DATE,
    ],
)
def test_parse_row_headers_failed(input_value):
    # then
    with pytest.raises(ValueError):
        # when
        row_headers_arg_parser.get_list(input_value)


from refinitiv.data.content import fundamental_and_reference
from tests.unit.conftest import StubSession, StubResponse, StubConfig
from tests.unit.content.fundamental.conftest import fund_and_ref_response


def test_workspace():
    from refinitiv.data.content import fundamental_and_reference

    assert hasattr(fundamental_and_reference, "Definition")


def test_attributes():
    # given
    excepted_attributes = [
        "universe",
        "fields",
        "parameters",
        "use_field_names_in_headers",
        "extended_params",
        "row_headers",
        "_kwargs",
        "_data_type",
        "_content_type",
        "_provider",
    ]

    # when
    definition = fundamental_and_reference.Definition(universe=[], fields=[])
    attributes = list(definition.__dict__.keys())

    # then
    assert attributes == excepted_attributes


def test_definition_fundamental_and_reference_repr():
    # given
    definition = fundamental_and_reference.Definition([], [])
    obj_id = hex(id(definition))
    expected_value = (
        f"<refinitiv.data.content.fundamental_and_reference.Definition object at {obj_id} {{universe='["
        f"]', fields='[]', parameters='None', row_headers='None'}}>"
    )

    # when
    testing_value = repr(definition)

    # then
    assert testing_value == expected_value


def test_platform_session_and_udf_layout_config_will_be_string():
    # given
    definition, session = create_definition_and_session("udf", SessionType.PLATFORM)

    # when
    definition.get_data(session)

    # then
    assert isinstance(definition._kwargs["layout"]["output"], str) is True
    assert definition._content_type == ContentType.DATA_GRID_RDP


def test_desktop_session_and_udf_layout_config_will_be_dict():
    # given
    definition, session = create_definition_and_session("udf", SessionType.DESKTOP)

    # when
    definition.get_data(session)

    # then
    assert isinstance(definition._kwargs["layout"], dict) is True
    assert definition._content_type == ContentType.DATA_GRID_UDF


def test_desktop_session_and_rdp_layout_config_will_be_dict():
    # given
    definition, session = create_definition_and_session("rdp", SessionType.DESKTOP)

    # when
    definition.get_data(session)

    # then
    assert isinstance(definition._kwargs["layout"]["output"], str) is True
    assert definition._content_type == ContentType.DATA_GRID_RDP


def test_platform_session_and_rdp_layout_config_will_be_string():
    # given
    definition, session = create_definition_and_session("rdp", SessionType.PLATFORM)

    # when
    definition.get_data(session)

    # then
    assert isinstance(definition._kwargs["layout"]["output"], str) is True
    assert definition._content_type == ContentType.DATA_GRID_RDP


def test_rdp_do_not_convert_number_to_date_if_field_have_function_and_date_into():
    response = StubResponse(
        # fmt: off
        {
            "links": {"count": 5}, "variability": "",
            "universe": [{
                "Instrument": "GOOG.O",
                "Company Common Name": "Alphabet Inc",
                "Organization PermID": "5030853586",
                "Reporting Currency": "USD"
            }, {
                "Instrument": "MSFT.O",
                "Company Common Name": "Microsoft Corp",
                "Organization PermID": "4295907168",
                "Reporting Currency": "USD"
            }, {
                "Instrument": "FB.O",
                "Company Common Name": "Meta Platforms Inc",
                "Organization PermID": "4297297477",
                "Reporting Currency": "USD"
            }, {
                "Instrument": "AMZN.O",
                "Company Common Name": "Amazon.com Inc",
                "Organization PermID": "4295905494",
                "Reporting Currency": "USD"
            }, {
                "Instrument": "TWTR.K",
                "Company Common Name": "Twitter Inc",
                "Organization PermID": "4296301199",
                "Reporting Currency": "USD"
            }],
            "data": [["GOOG.O", 92.168064516129], ["MSFT.O", 241.904193548387],
                     ["FB.O", 121.724838709677], ["AMZN.O", 88.6803225806452],
                     ["TWTR.K", 53.7]],
            "messages": {
                "codes": [[-1, -1], [-1, -1], [-1, -1], [-1, -1], [-1, -1]],
                "descriptions": [{"code": -1, "description": "ok"}]
            },
            "headers": [{
                "name": "instrument", "title": "Instrument",
                "type": "string",
                "description": "The requested Instrument as defined by the user."
            }, {
                "name": "AVG(TR.Close(Edate=0, SDate = -30))",
                "title": "AVG(TR.Close(Edate=0, SDate = -30))",
                "type": "number", "decimalChar": ".", "description": None
            }]
        }
        # fmt: on
    )
    session = StubSession(is_open=True, response=response)
    session.config.set_param("apis.data.datagrid.underlying-platform", "rdp")
    field = "AVG(TR.Close(Edate=0, SDate = -30))"
    definition = fundamental_and_reference.Definition(
        universe=["GOOG.O", "MSFT.O", "FB.O", "AMZN.O", "TWTR.K"], fields=[field]
    )
    response = definition.get_data(session)

    assert response.data.df[field][0] == 92.168064516129


def test_udf_do_not_convert_number_to_date_if_field_have_function_and_date_into():
    response = StubResponse(
        # fmt: off
        {
            "responses": [{
                "columnHeadersCount": 1,
                "data": [["GOOG.O", 91.8683870967742], ["MSFT.O", 241.427096774194],
                         ["FB.O", 122.281612903226], ["AMZN.O", 88.6645161290323],
                         ["TWTR.K", 53.7]],
                "headerOrientation": "horizontal",
                "headers": [
                    [{"displayName": "Instrument"}, {
                        "displayName": "AVG(TR.Close(Edate=0, SDate = -30))",
                        "field": "AVG(TR.Close(Edate=0, SDate = -30))"
                    }]], "rowHeadersCount": 1, "totalColumnsCount": 2,
                "totalRowsCount": 6
            }]
        }
        # fmt: on
    )
    session = StubSession(is_open=True, response=response)
    session.type = SessionType.DESKTOP
    session.config.set_param("apis.data.datagrid.underlying-platform", "udf")
    field = "AVG(TR.Close(Edate=0, SDate = -30))"
    instrument = "GOOG.O"
    definition = fundamental_and_reference.Definition(
        universe=[instrument, "MSFT.O", "FB.O", "AMZN.O", "TWTR.K"], fields=[field]
    )
    response = definition.get_data(session)

    assert response.data.df[field][0] == 91.8683870967742
    assert response.data.df["Instrument"][0] == instrument



from refinitiv.data.content.ipa._curves._cross_currency_curves import (
    RiskType,
    MainConstituentAssetClass,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions._get import (
    CrossCurrencyCurveDefinitionKeys,
    GetRequest,
)


def test_cross_currency_curve_definition_keys():
    # given
    curve_definition = CrossCurrencyCurveDefinitionKeys(
        main_constituent_asset_class=MainConstituentAssetClass.FX_FORWARD,
        risk_type=RiskType.CROSS_CURRENCY,
        base_currency="string",
        base_index_name="string",
        id="string",
        is_non_deliverable=True,
        name="string",
        quoted_currency="string",
        quoted_index_name="string",
        source="string",
    )

    # when
    testing_obj = GetRequest(curve_definition=curve_definition)

    # then
    assert testing_obj.get_dict() == {"curveDefinition": curve_definition}

    assert testing_obj.curve_definition == curve_definition



from typing import Optional


from ._curves._credit_curve_definition import CreditCurveDefinition
from ._curves._credit_curve_parameters import CreditCurveParameters


CurveDefinition = Optional[CreditCurveDefinition]
CurveParameters = Optional[CreditCurveParameters]



import json
from inspect import signature

import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.delivery._data._data_provider_factory import make_provider
from tests.unit.conftest import StubSession
from tests.unit.conftest import (
    remove_dunder_methods,
    remove_private_attributes,
    has_property_names_in_class,
    get_property_names,
)


def test_ipa_bond_definition_attributes():
    expected_attributes = ["get_data", "get_data_async", "get_stream"]
    testing_attributes = dir(rdf.bond.Definition)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes


def test_ipa_bond_attributes():
    expected_attributes = [
        "AdjustInterestToPaymentDate",
        "AmortizationFrequency",
        "AmortizationItem",
        "AmortizationType",
        "BenchmarkYieldSelectionMode",
        "BondRoundingParameters",
        "BusinessDayConvention",
        "CreditSpreadType",
        "DateRollingConvention",
        "DayCountBasis",
        "Definition",
        "Direction",
        "DividendType",
        "Frequency",
        "IndexAverageMethod",
        "IndexCompoundingMethod",
        "InflationMode",
        "InterestType",
        "PriceSide",
        "PricingParameters",
        "ProjectedIndexCalculationMethod",
        "QuoteFallbackLogic",
        "RedemptionDateType",
        "Rounding",
        "RoundingType",
        "StubRule",
        "VolatilityTermStructureType",
        "VolatilityType",
        "YieldType",
    ]
    testing_attributes = dir(rdf.bond)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes, testing_attributes


@pytest.mark.parametrize(
    argnames="input_data",
    ids=[
        "Definition",
        "PricingParameters",
        "BondRoundingParameters",
        "AmortizationItem",
    ],
    argvalues=[
        (
            rdf.bond._bond_definition.BondInstrumentDefinition,
            {
                "instrument_code": "instrument_code",
                "instrument_tag": "instrument_tag",
                "end_date": "end_date",
                "direction": rdf.bond.Direction.PAID,
                "interest_type": rdf.bond.InterestType.FIXED,
                "notional_ccy": "notional_ccy",
                "notional_amount": 7.7,
                "fixed_rate_percent": 8.8,
                "spread_bp": 9.9,
                "interest_payment_frequency": rdf.bond.Frequency.EVERY14_DAYS,
                "interest_calculation_method": rdf.bond.DayCountBasis.DCB_30_360_ISDA,
                "accrued_calculation_method": rdf.bond.DayCountBasis.DCB_30_360,
                "payment_business_day_convention": rdf.bond.BusinessDayConvention.MODIFIED_FOLLOWING,
                "payment_roll_convention": rdf.bond.DateRollingConvention.LAST28,
                "index_reset_frequency": rdf.bond.Frequency.ANNUAL,
                "index_fixing_lag": 16,
                "first_regular_payment_date": "first_regular_payment_date",
                "last_regular_payment_date": "last_regular_payment_date",
                "amortization_schedule": [rdf.bond.AmortizationItem()],
                "payment_business_days": "payment_business_days",
                "adjust_interest_to_payment_date": rdf.bond.AdjustInterestToPaymentDate.ADJUSTED,
                "index_compounding_method": rdf.bond.IndexCompoundingMethod.MEXICAN_COMPOUNDED,
                "interest_payment_delay": 23,
                "stub_rule": rdf.bond.StubRule.ISSUE,
                "issue_date": "issue_date",
                "index_average_method": rdf.bond.IndexAverageMethod.ARITHMETIC_AVERAGE,
                "first_accrual_date": "str",
                "floor_strike_percent": 10.10,
                "index_fixing_ric": "index_fixing_ric",
                "is_perpetual": True,
                "template": "str",
            },
        ),
        (
            rdf.bond.PricingParameters,
            {
                "trade_date": "trade_date",
                "benchmark_yield_selection_mode": rdf.bond.BenchmarkYieldSelectionMode.INTERPOLATE,
                "credit_spread_type": rdf.bond.CreditSpreadType.TERM_STRUCTURE,
                "dividend_type": rdf.bond.DividendType.FUTURES,
                "volatility_term_structure_type": rdf.bond.VolatilityTermStructureType.HISTORICAL,
                "volatility_type": rdf.bond.VolatilityType.FLAT,
                "bond_recovery_rate_percent": 123.123,
                "cds_recovery_rate_percent": 123.123,
                "dividend_yield_percent": 123.123,
                "flat_credit_spread_bp": 123.123,
                "flat_credit_spread_tenor": "flat_credit_spread_tenor",
                "fx_stock_correlation": 123.123,
                "fx_volatility_percent": 123.123,
                "fx_volatility_tenor": "fx_volatility_tenor",
                "stock_borrow_rate_percent": 123.123,
                "stock_flat_volatility_percent": 123.123,
                "stock_flat_volatility_tenor": "stock_flat_volatility_tenor",
                "stock_price_on_default": 123.123,
                "market_data_date": "market_data_date",
                "price_side": rdf.bond.PriceSide.MID,
                "projected_index_calculation_method": rdf.bond.ProjectedIndexCalculationMethod.CONSTANT_INDEX,
                "redemption_date_type": rdf.bond.RedemptionDateType.REDEMPTION_AT_BEST_DATE,
                "rounding_parameters": rdf.bond.BondRoundingParameters(),
                "yield_type": rdf.bond.YieldType.DISCOUNT_ACTUAL_360,
                "adjusted_clean_price": 9.9,
                "adjusted_dirty_price": 10.10,
                "adjusted_yield_percent": 11.11,
                "apply_tax_to_full_pricing": True,
                "asset_swap_spread_bp": 13.13,
                "benchmark_at_issue_price": 14.14,
                "benchmark_at_issue_ric": "benchmark_at_issue_ric",
                "benchmark_at_issue_spread_bp": 16.16,
                "benchmark_at_issue_yield_percent": 17.17,
                "benchmark_at_redemption_price": 18.18,
                "benchmark_at_redemption_spread_bp": 19.19,
                "benchmark_at_redemption_yield_percent": 20.20,
                "cash_amount": 21.21,
                "clean_price": 22.22,
                "concession_fee": 23.23,
                "current_yield_percent": 24.24,
                "dirty_price": 25.25,
                "discount_margin_bp": 26.26,
                "discount_percent": 27.27,
                "edsf_benchmark_curve_yield_percent": 28.28,
                "edsf_spread_bp": 29.29,
                "efp_benchmark_price": 30.30,
                "efp_benchmark_ric": "efp_benchmark_ric",
                "efp_benchmark_yield_percent": 32.32,
                "efp_spread_bp": 33.33,
                "gov_country_benchmark_curve_price": 34.34,
                "gov_country_benchmark_curve_yield_percent": 35.35,
                "gov_country_spread_bp": 36.36,
                "government_benchmark_curve_price": 37.37,
                "government_benchmark_curve_yield_percent": 38.38,
                "government_spread_bp": 39.39,
                "issuer_benchmark_curve_yield_percent": 40.40,
                "issuer_spread_bp": 41.41,
                "market_value_in_deal_ccy": 42.42,
                "market_value_in_report_ccy": 43.43,
                "net_price": 44.44,
                "neutral_yield_percent": 45.45,
                "option_adjusted_spread_bp": 48.48,
                "price": 49.49,
                "quoted_price": 50.50,
                "rating_benchmark_curve_yield_percent": 51.51,
                "rating_spread_bp": 52.52,
                "redemption_date": "redemption_date",
                "sector_rating_benchmark_curve_yield_percent": 54.54,
                "sector_rating_spread_bp": 55.55,
                "settlement_convention": "settlement_convention",
                "simple_margin_bp": 57.57,
                "strip_yield_percent": 58.58,
                "swap_benchmark_curve_yield_percent": 59.59,
                "swap_spread_bp": 60.60,
                "tax_on_capital_gain_percent": 61.61,
                "tax_on_coupon_percent": 62.62,
                "tax_on_price_percent": 63.63,
                "tax_on_yield_percent": 64.64,
                "user_defined_benchmark_price": 66.66,
                "user_defined_benchmark_yield_percent": 67.67,
                "user_defined_spread_bp": 68.68,
                "valuation_date": "valuation_date",
                "yield_percent": 70.70,
                "z_spread_bp": 71.71,
                "quote_fallback_logic": rdf.bond.QuoteFallbackLogic.BEST_FIELD,
                "fx_price_side": rdf.bond.PriceSide.BID,
                "ois_zc_benchmark_curve_yield_percent": 72.72,
                "ois_zc_spread_bp": 73.73,
                "use_settlement_date_from_quote": True,
                "compute_cash_flow_with_report_ccy": True,
                "inflation_mode": rdf.bond.InflationMode.DEFAULT,
                "next_coupon_rate_percent": 74.74,
                "is_coupon_payment_adjustedfor_leap_year": True,
                "report_ccy": "report_ccy",
                "compute_cash_flow_from_issue_date": True,
                "projected_index_percent": 75.75,
            },
        ),
        (
            rdf.bond.BondRoundingParameters,
            {
                "accrued_rounding": rdf.bond.Rounding.FOUR,
                "accrued_rounding_type": rdf.bond.RoundingType.FACE_NEAR,
                "price_rounding": rdf.bond.Rounding.FIVE,
                "price_rounding_type": rdf.bond.RoundingType.FACE_DOWN,
                "spread_rounding": rdf.bond.Rounding.EIGHT,
                "spread_rounding_type": rdf.bond.RoundingType.CEIL,
                "yield_rounding": rdf.bond.Rounding.DEFAULT,
                "yield_rounding_type": rdf.bond.RoundingType.FACE_UP,
            },
        ),
        (
            rdf.bond.AmortizationItem,
            {
                "start_date": "start_date",
                "end_date": "end_date",
                "amortization_frequency": rdf.bond.AmortizationFrequency.ONCE,
                "amortization_type": rdf.bond.AmortizationType.LINEAR,
                "remaining_notional": 1.1,
                "amount": 2.2,
            },
        ),
    ],
)
def test_parameter(input_data):
    cls, kwargs = input_data
    args_names = list(kwargs.keys())
    inst = cls(**kwargs)

    s = signature(cls.__init__)
    # +1 for (self)
    assert len(s.parameters) == (
        len(args_names) + 1
    ), f"csl={cls}{str(set(get_property_names(cls)) - set(args_names))}"

    assert has_property_names_in_class(cls, args_names), set(args_names) - set(
        get_property_names(cls)
    )

    for k, v in kwargs.items():
        attr = getattr(inst, k)
        assert attr == v, k

    try:
        json_data = inst.get_dict()
        json.dumps(json_data)
    except Exception as e:
        assert False, str(e)


def test_bond_definition_will_raise_exception_if_pass_not_str():
    with pytest.raises(ValueError, match="Invalid type of instrument_code"):
        rdf.bond.Definition(["one", "two"])


def test_none_value_is_not_set_instrument_code():
    # given
    bond_definition = rdf.bond.Definition(
        issue_date="2002-02-28",
        end_date="2032-02-28",
        notional_ccy="USD",
        interest_payment_frequency="Annual",
        fixed_rate_percent=7,
        interest_calculation_method=rdf.bond.DayCountBasis.DCB_ACTUAL_ACTUAL,
    )
    definition = bond_definition._kwargs["definition"]

    # when
    testing_value = definition.instrument_code

    # then
    assert testing_value is None


def test_request_body():
    # given
    expected_body = {
        "fields": ["1", "2"],
        "universe": [
            {
                "instrumentDefinition": {
                    "endDate": "2032-02-28",
                    "fixedRatePercent": 7,
                    "interestCalculationMethod": "Dcb_Actual_Actual",
                    "interestPaymentFrequency": "Annual",
                    "issueDate": "2002-02-28",
                    "notionalCcy": "USD",
                },
                "extended_params": "extended_params",
                "instrumentType": "Bond",
                "pricingParameters": {"tradeDate": "trade_date"},
            }
        ],
    }
    definition = rdf.bond.Definition(
        issue_date="2002-02-28",
        end_date="2032-02-28",
        notional_ccy="USD",
        interest_payment_frequency="Annual",
        fixed_rate_percent=7,
        interest_calculation_method=rdf.bond.DayCountBasis.DCB_ACTUAL_ACTUAL,
        extended_params={"extended_params": "extended_params"},
        pricing_parameters=rdf.bond.PricingParameters(trade_date="trade_date"),
        fields=["1", "2"],
    )

    # when
    content_type = definition._kwargs["__content_type__"]
    provider = make_provider(content_type)
    request = provider.request.create(StubSession(), "url", **definition._kwargs)

    # then
    testing_body = request.json
    assert testing_body == expected_body, expected_body



from enum import Enum, unique
from typing import Union, List, Optional

from ..._tools import make_enum_arg_parser


# --------------------------------------------------------------------------------------
#   EventTypes
# --------------------------------------------------------------------------------------


@unique
class EventTypes(Enum):
    """
    The list of market events (comma delimiter), supported event types are trade,
    quote and correction.
    Note: Currently support only single event type.
        If request with multiple event types,
        the backend will pick up the first event type to proceed.
    """

    TRADE = "trade"
    QUOTE = "quote"
    CORRECTION = "correction"


OptEventTypes = Optional[Union[str, List[str], EventTypes, List[EventTypes]]]
event_types_arg_parser = make_enum_arg_parser(EventTypes)


# --------------------------------------------------------------------------------------
#   Adjustments
# --------------------------------------------------------------------------------------


@unique
class Adjustments(Enum):
    """
    The list of adjustment types (comma delimiter) that tells the system whether
     to apply or not apply CORAX (Corporate Actions) events or
     exchange/manual corrections to historical time series data.

     The supported values of adjustments :

        UNADJUSTED - Not apply both exchange/manual corrections and CORAX
        EXCHANGE_CORRECTION - Apply exchange correction adjustment to historical pricing
        MANUAL_CORRECTION - Apply manual correction adjustment to historical pricing
                            i.e. annotations made by content analysts
        CCH - Apply Capital Change adjustment to historical Pricing due
              to Corporate Actions e.g. stock split
        CRE - Apply Currency Redenomination adjustment
              when there is redenomination of currency
        RPO - Apply Reuters Price Only adjustment
              to adjust historical price only not volume
        RTS - Apply Reuters TimeSeries adjustment
              to adjust both historical price and volume
        QUALIFIERS - Apply price or volume adjustment
              to historical pricing according to trade/quote qualifier
              summarization actions
    """

    UNADJUSTED = "unadjusted"
    EXCHANGE_CORRECTION = "exchangeCorrection"
    MANUAL_CORRECTION = "manualCorrection"
    CCH = "CCH"
    CRE = "CRE"
    RPO = "RPO"
    RTS = "RTS"
    QUALIFIERS = "qualifiers"


OptAdjustments = Optional[Union[str, List[str], Adjustments, List[Adjustments]]]
adjustments_arg_parser = make_enum_arg_parser(Adjustments)


# --------------------------------------------------------------------------------------
#   MarketSession
# --------------------------------------------------------------------------------------


@unique
class MarketSession(Enum):
    """
    The marketsession parameter represents a list of interested official durations
        in which trade and quote activities occur for a particular universe.

    The supported values of marketsession :

        PRE - specifies that data returned
              should include data during pre-market session
        NORMAL - specifies that data returned
                 should include data during normal market session
        POST - specifies that data returned
               should include data during post-market session
    """

    PRE = "pre"
    NORMAL = "normal"
    POST = "post"


OptMarketSession = Optional[Union[str, List[str], MarketSession, List[MarketSession]]]
market_sessions_arg_parser = make_enum_arg_parser(MarketSession)



import allure
import pytest

from refinitiv.data.content.filings import search
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.filings.conftest import (
    check_file_downloaded_response,
    check_file_is_downloaded,
    get_callback,
)
from tests.integration.helpers import (
    get_async_response_from_definition,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
)

query = (
    '{  FinancialFiling(filter: {AND: [{FilingDocument: {DocumentSummary: {FeedName: {EQ: "Edgar"}}}}, '
    '{FilingDocument: {DocumentSummary: {FormType: {EQ: "10-Q"}}}}, {FilingDocument: {DocumentSummary: {'
    'FilingDate: {BETWN: {FROM: "2020-10-01T00:00:00Z", TO: "2020-12-31T00:00:00Z"}}}}}]}, sort: {FilingDocument: '
    "{DocumentSummary: {FilingDate: DESC}}}, limit: 3) {    _metadata {      totalCount    }    FilingDocument { "
    "     Identifiers {        Dcn      }      DocId      FinancialFilingId      DocumentSummary {        "
    "DocumentTitle        FeedName        FormType        HighLevelCategory        MidLevelCategory        "
    "FilingDate        SecAccessionNumber        SizeInBytes          }  FilesMetaData {        FileName        "
    "MimeType      }    }  }} "
)


@allure.suite("Filings Search")
@allure.feature("Filings Search")
@allure.severity(allure.severity_level.CRITICAL)
class TestFilingsSearch:
    @allure.title("Create filings search definition with valid params and get data")
    @pytest.mark.parametrize(
        "expected_titles",
        [
            [
                "DocumentTitle",
                "Filename",
                "MimeType",
                "Dcn",
                "DocId",
                "FinancialFilingId",
            ],
        ],
    )
    @pytest.mark.caseid("C50564778")
    @pytest.mark.smoke
    def test_create_filings_search_definition_with_valid_params_and_get_data(
        self, open_platform_session, expected_titles
    ):
        response = search.Definition(query).get_data()

        file = response.data.files
        file_download_response = file[0].download(path="./")
        expected_file_name = response.data.df["Filename"][0]

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_contains_columns_names(response, expected_titles)
        check_non_empty_response_data(response)

        check_file_downloaded_response(file_download_response, expected_file_name)
        check_file_is_downloaded(expected_file_name)

    @allure.title("Create filings search definition and download all files")
    @pytest.mark.caseid("C50564779")
    @pytest.mark.smoke
    def test_create_filings_search_definition_and_download_all_files(
        self, open_platform_session
    ):
        response = search.Definition(query).get_data()

        files = response.data.files
        files.download()

        expected_file_name = response.data.df["Filename"].values
        for filename in expected_file_name:
            check_file_is_downloaded(filename)

    @allure.title(
        "Create filings sync search definition and download all files asynchronously"
    )
    @pytest.mark.parametrize(
        "expected_titles",
        [
            [
                "DocumentTitle",
                "Filename",
                "MimeType",
                "Dcn",
                "DocId",
                "FinancialFilingId",
            ]
        ],
    )
    @pytest.mark.caseid("C50564780")
    @pytest.mark.smoke
    async def test_create_filings_sync_search_definition_and_download_all_files_asynchronously(
            self, expected_titles, open_platform_session_async
    ):
        response = await get_async_response_from_definition(search.Definition(query))

        files = response.data.files
        await files.download_async(callback=get_callback)

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_contains_columns_names(response, expected_titles)
        expected_file_name = response.data.df["Filename"].values
        for filename in expected_file_name:
            check_file_is_downloaded(filename)

    @allure.title(
        "Create sync filings search definition and download first file asynchronously"
    )
    @pytest.mark.parametrize(
        "expected_titles",
        [
            [
                "DocumentTitle",
                "Filename",
                "MimeType",
                "Dcn",
                "DocId",
                "FinancialFilingId",
            ]
        ],
    )
    @pytest.mark.caseid("C50645502")
    @pytest.mark.smoke
    async def test_create_filings_search_definition_and_download_first_files_asynchronously(
        self, expected_titles, open_platform_session_async
    ):
        response = search.Definition(query).get_data()

        files = response.data.files
        await files[0].download_async(callback=get_callback)

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_contains_columns_names(response, expected_titles)
        expected_file_name = response.data.df["Filename"].values
        for filename in expected_file_name:
            check_file_is_downloaded(filename)


from typing import Union, TYPE_CHECKING

from ._custom_instruments_data_provider import get_content_type_by_interval
from .._content_data import Data
from .._content_provider_layer import ContentUsageLoggerMixin
from .._intervals import DayIntervalType, get_day_interval_type, Intervals
from ..._tools import (
    validate_types,
    custom_insts_historical_universe_parser,
    try_copy_to_list,
)
from ...delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ..._types import StrStrings, OptDateTime, OptInt, ExtendedParams


class Definition(
    ContentUsageLoggerMixin[BaseResponse[Data]], DataProviderLayer[BaseResponse[Data]]
):
    """
    Summary line of this class that defines parameters for requesting summaries from custom instruments

    Parameters
    ----------
    universe : str or list
        The Id or Symbol of custom instrument to operate on
    interval : str or Intervals, optional
        The consolidation interval in ISO8601
    start : str or date or datetime or timedelta, optional
        The start date and timestamp of the query in ISO8601 with UTC only
    end : str or date or datetime or timedelta, optional
        The end date and timestamp of the query in ISO8601 with UTC only
    count : int, optional
        The maximum number of data returned. Values range: 1 - 10000
    fields : list, optional
        The list of fields that are to be returned in the response
    extended_params : dict, optional
        If necessary other parameters

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments import summaries
    >>> definition_summaries = summaries.Definition("VOD.L")
    >>> response = definition_summaries.get_data()
    """

    _USAGE_CLS_NAME = "CustomInstruments.SummariesDefinition"

    def __init__(
        self,
        universe: "StrStrings",
        interval: Union[str, Intervals] = None,
        start: "OptDateTime" = None,
        end: "OptDateTime" = None,
        count: "OptInt" = None,
        fields: "StrStrings" = None,
        extended_params: "ExtendedParams" = None,
    ) -> None:
        day_interval_type = get_day_interval_type(interval or DayIntervalType.INTER)
        content_type = get_content_type_by_interval(day_interval_type)
        validate_types(count, [int, type(None)], "count")

        fields = try_copy_to_list(fields)
        universe = try_copy_to_list(universe)
        universe = custom_insts_historical_universe_parser.get_list(universe)

        super().__init__(
            data_type=content_type,
            universe=universe,
            interval=interval,
            start=start,
            end=end,
            count=count,
            fields=fields,
            extended_params=extended_params,
        )



# coding: utf8


from enum import Enum, unique


@unique
class FundingSpreadMethod(Enum):
    CDS_SPREAD = "CDSSpread"
    LIQUIDITY_SPREAD = "LiquiditySpread"
    PPZ_SPREAD = "PPZSpread"
    PARALLEL_CURVE_SHIFT = "ParallelCurveShift"
    RISKY_D_FS = "RiskyDFs"
    USE_CDS_SPREADS = "UseCDSSpreads"
    USE_Z_SPREAD = "UseZSpread"
    Z_SPREAD = "ZSpread"



import refinitiv.data.content.ipa.financial_contracts as rdf


def cap_floor_on_cms_definition():
    definition = rdf.cap_floor.Definition(
        instrument_tag="CapOnCms",
        stub_rule=rdf.cap_floor.StubRule.MATURITY,
        notional_ccy="USD",
        start_date="2018-06-15",
        end_date="2022-06-15",
        notional_amount=1000000,
        index_name="Composite",
        index_tenor="5Y",
        cms_template="USD_SB3L",
        interest_calculation_method="Dcb_Actual_360",
        interest_payment_frequency=rdf.cap_floor.Frequency.QUARTERLY,
        buy_sell=rdf.cap_floor.BuySell.BUY,
        cap_strike_percent=1,
        pricing_parameters=rdf.cap_floor.PricingParameters(
            skip_first_cap_floorlet=True, valuation_date="2020-02-07"
        ),
        fields=[
            "InstrumentTag",
            "InstrumentDescription",
            "StartDate",
            "EndDate",
            "MarketDataDate",
            "InterestPaymentFrequency",
            "IndexRic",
            "CapStrikePercent",
            "FloorStrikePercent",
            "NotionalCcy",
            "NotionalAmount",
            "PremiumBp",
            "PremiumPercent",
            "MarketValueInDealCcy",
            "MarketValueInReportCcy",
            "ErrorMessage",
        ],
    )
    return definition


def cap_floor_coller_pos_definition():
    definition = rdf.cap_floor.Definition(
        notional_ccy="EUR",
        start_date="2018-02-11",
        index_reset_type=rdf.cap_floor.IndexResetType.IN_ARREARS,
        tenor="5Y",
        buy_sell=rdf.cap_floor.BuySell.BUY,
        notional_amount=10000000,
        interest_payment_frequency=rdf.cap_floor.Frequency.QUARTERLY,
        cap_strike_percent=1,
        floor_strike_percent=-1,
        pricing_parameters=rdf.cap_floor.PricingParameters(
            skip_first_cap_floorlet=False, valuation_date="2020-02-07"
        ),
        fields=[
            "InstrumentTag",
            "InstrumentDescription",
            "OptionType",
            "StartDate",
            "EndDate",
            "MarketDataDate",
            "InterestPaymentFrequency",
            "IndexResetFrequency",
            "IndexRic",
            "CapStrikePercent",
            "FloorStrikePercent",
            "NotionalCcy",
            "NotionalAmount",
            "PremiumBp",
            "PremiumPercent",
            "MarketValueInDealCcy",
            "MarketValueInReportCcy",
            "ImpliedVolatilityBp",
            "ImpliedVolatilityPercent",
            "Legs",
            "DeltaPercent",
            "ForwardDeltaPercent",
            "DeltaAmountInDealCcy",
            "DeltaAmountInReportCcy",
            "VegaAmountInDealCcy",
            "VegaAmountInReportCcy",
            "ThetaAmountInDealCcy",
            "ThetaAmountInReportCcy",
            "ErrorCode",
            "ErrorMessage",
        ],
    )
    return definition


def cap_floor_amortized_definition():
    definition = rdf.cap_floor.Definition(
        instrument_tag="CapGBP",
        notional_ccy="GBP",
        start_date="2021-06-11",
        amortization_schedule=[
            rdf.cap_floor.AmortizationItem(
                start_date="2023-06-12",
                end_date="2024-06-12",
                amount=100000,
                amortization_type="Schedule",
                amortization_frequency=rdf.cap_floor.AmortizationFrequency.EVERY_COUPON,
            ),
            rdf.cap_floor.AmortizationItem(
                start_date="2024-06-11",
                end_date="2025-06-11",
                amount=-100000,
                amortization_type="Schedule",
                amortization_frequency=rdf.cap_floor.AmortizationFrequency.EVERY_COUPON,
            ),
        ],
        tenor="10Y",
        buy_sell="Buy",
        notional_amount=10000000,
        interest_payment_frequency="Quarterly",
        cap_strike_percent=0.25,
        pricing_parameters=rdf.cap_floor.PricingParameters(
            skip_first_cap_floorlet=False,
            valuation_date="2021-10-07",
            market_data_date="2020-10-07",
        ),
        fields=[
            "InstrumentTag",
            "InstrumentDescription",
            "FixedRate",
            "MarketDataDate",
            "MarketValueInDealCcy",
            "MarketValueInReportCcy",
            "ErrorMessage",
            "MarketDataDate",
        ],
    )
    return definition


def invalid_cap_floor_definition():
    definition = rdf.cap_floor.Definition(
        notional_ccy="INVAL",
        start_date="2019-02-11",
        amortization_schedule=[
            rdf.cap_floor.AmortizationItem(
                start_date="2021-02-12",
                end_date="2021-02-12",
                amount=100000,
                amortization_type="Schedule",
            ),
            rdf.cap_floor.AmortizationItem(
                start_date="2021-02-11",
                end_date="2021-02-11",
                amount=-100000,
                amortization_type="Schedule",
            ),
        ],
        tenor="5Y",
        buy_sell="Sell",
        notional_amount=10000000,
        interest_payment_frequency="Monthly",
        cap_strike_percent=1,
        pricing_parameters=rdf.cap_floor.PricingParameters(
            skip_first_cap_floorlet=True, valuation_date="2020-02-07"
        ),
        fields=[
            "InstrumentTag",
            "InstrumentDescription",
            "FixedRate",
            "MarketValueInDealCcy",
            "MarketValueInReportCcy",
            "ErrorMessage",
            "MarketDataDate",
        ],
    )
    return definition


cap_floor_universe = {
    "instrumentType": "CapFloor",
    "instrumentDefinition": {
        "instrumentTag": "CapGBP",
        "startDate": "2021-06-11",
        "tenor": "10Y",
        "notionalCcy": "GBP",
        "notionalAmount": 10000000,
        "interestPaymentFrequency": "Quarterly",
        "amortizationSchedule": [
            {
                "startDate": "2023-06-12",
                "endDate": "2024-06-12",
                "amortizationFrequency": "EveryCoupon",
                "amortizationType": "Schedule",
                "amount": 100000,
            },
            {
                "startDate": "2024-06-11",
                "endDate": "2025-06-11",
                "amortizationFrequency": "EveryCoupon",
                "amortizationType": "Schedule",
                "amount": -100000,
            },
        ],
        "buySell": "Buy",
        "capStrikePercent": 0.25,
    },
    "pricingParameters": {
        "marketDataDate": "2020-10-07",
        "skipFirstCapFloorlet": False,
        "valuationDate": "2021-10-07",
    },
}



from typing import TYPE_CHECKING, Any, Generic, TypeVar, Union

if TYPE_CHECKING:
    from . import Session


# --------------------------------------
# PUBLIC API
# --------------------------------------


def set_default(session: "Session") -> None:
    """
    Set the default session.

    Parameters
    ----------
    session: Session
        The logical session to the data platform.

    Returns
    -------
    """
    _rd_default_session_manager.set_default_session(session)


def get_default() -> "Session":
    """
    Gets the default session.

    Returns
    -------
    Session
    """
    return _rd_default_session_manager.try_get_default_session()


# --------------------------------------
# PRIVATE API
# --------------------------------------
class SessionManager:
    def __init__(self, arg: Union["Session", "Wrapper"] = None) -> None:
        if not isinstance(arg, Wrapper):
            arg = Wrapper(arg)
        self._session_wrapper: Wrapper = arg

    def close_default_session(self):
        if self._session_wrapper.get() is not None:
            self._session_wrapper.get().close()

    def clear_default_session(self):
        self.close_default_session()
        self._session_wrapper.set(None)

    def has_session(self):
        return self._session_wrapper.get() is not None


class EikonDefaultSessionManager(SessionManager):
    def get_default_session(self, app_key=None):
        from ...session.desktop import Definition

        is_create_session = False
        if self._session_wrapper.get() is None:
            if app_key is None:
                return None
            is_create_session = True

        elif app_key is not None and app_key != self._session_wrapper.get().app_key:
            self._session_wrapper.get().close()
            is_create_session = True

        if is_create_session:
            definition = Definition(app_key=app_key)
            self._session_wrapper.set(definition.get_session())

        return self._session_wrapper.get()

    def set_default_session(self, session):
        from . import Session

        if isinstance(session, Session):
            self._session_wrapper.set(session)


class RDDefaultSessionManager(SessionManager):
    def try_get_default_session(self):
        """
        Returns default session or raise exception if user didn't set default session.

        Raises
        ------
        Exception
            If user didn't set default session

        Returns
        ------
        Session object
        """
        if not self._session_wrapper.get():
            raise AttributeError(
                "No default session created yet. Please create a session first!"
            )
        return self._session_wrapper.get()

    def set_default_session(self, session) -> None:
        """
        Set default session.

        Parameters
        ----------
            session: Session
                session object

        Raises
        ----------
        Exception
            If user provided invalid data type
        """
        from . import Session

        if session is None:
            return self.clear_default_session()

        if not isinstance(session, Session):
            raise TypeError("Invalid argument")

        self._session_wrapper.set(session)

    def set_session(self, session):
        self._session_wrapper.set(session)


T = TypeVar("T")


class Wrapper(Generic[T]):
    def __init__(self, wrapped: T = None):
        self._wrapped: T = wrapped

    def get(self) -> T:
        return self._wrapped

    def set(self, value: T):
        self._wrapped = value


def get_valid_session(session: Any) -> "Session":
    if session is None:
        session = get_default()
    if session is None:
        raise AttributeError("A Session must be started")
    return session


_session_wrapper = Wrapper()
_rd_default_session_manager = RDDefaultSessionManager(_session_wrapper)
_eikon_default_session_manager = EikonDefaultSessionManager(_session_wrapper)



import json
from typing import Callable, List, Dict

import websocket


class WebSocketClient:
    def __init__(
        self, on_message: Callable[[List[Dict]], None], *args, **kwargs
    ) -> None:
        # set global timeout for WebSocketApp according documentation
        # https://websocket-client.readthedocs.io/en/latest/examples.html#setting-timeout-value
        websocket.setdefaulttimeout(5)

        self._on_message = on_message
        self.ws = websocket.WebSocketApp(
            on_message=self._on_ws_message, *args, **kwargs
        )

    def _on_ws_message(self, ws: websocket.WebSocketApp, s: str) -> None:
        try:
            messages = json.loads(s)
        except UnicodeDecodeError:
            messages = "".join(map(chr, bytearray(s)))
            messages = json.loads(messages)
        self._on_message(messages)

    def send(self, data: dict) -> None:
        self.ws.send(json.dumps(data))

    def run_forever(self, *args, **kwargs) -> None:
        self.ws.run_forever(*args, **kwargs)

    def close(self, **kwargs) -> None:
        self.ws.close(**kwargs)



from datetime import datetime

from refinitiv.data import OpenState


def assert_session_open_state(session):
    assert session.open_state == OpenState.Opened
    session.close()
    assert session.open_state == OpenState.Closed


def on_state(state, message, session, event_list=None):
    current_time = datetime.now().time()
    print("----------------------------------------------------------")
    print(
        ">>> {} received at {} session state is {}".format(
            message, current_time, state
        )
    )
    if event_list is not None:
        event_list.append((state, message))


def on_event(event_code, message, session, event_list=None):
    current_time = datetime.now().time()
    print("----------------------------------------------------------")
    print(
        ">>> {} received at {} event code is {}".format(
            message, current_time, event_code
        )
    )
    if event_list is not None:
        event_list.append((event_code, message))



from datetime import datetime, timedelta

import allure
import pytest

from refinitiv.data.content.ipa.dates_and_calendars import is_working_day
from refinitiv.data.content.ipa.dates_and_calendars.is_working_day import HolidayOutputs
from refinitiv.data.content.ipa.dates_and_calendars.is_working_day._is_working_day_data_provider import (
    WorkingDay,
)
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.helpers import (
    check_response_status,
    get_async_response_from_definitions,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_extended_params_were_sent_in_request,
)

first_definition = is_working_day.Definition(
    tag="my request",
    date="2022-01-01",
    calendars=["BAR", "KOR", "JAP"],
    currencies=["USD"],
    holiday_outputs=[HolidayOutputs.DATE, HolidayOutputs.NAMES],
)

second_definition = is_working_day.Definition(
    tag="my request",
    date="2021-12-25",
    calendars=["BAR"],
    currencies=["EUR"],
    holiday_outputs=[HolidayOutputs.DATE, HolidayOutputs.NAMES],
)

invalid_definition = is_working_day.Definition(
    date="2021.04.24",
    calendars=["JAP11"],
    holiday_outputs=[HolidayOutputs.DATE, HolidayOutputs.NAMES],
)


@allure.suite("Dates And Calendars")
@allure.feature("Content object - Is Working Day")
@allure.severity(allure.severity_level.CRITICAL)
class TestIsWorkingDay:
    @allure.title("Create a is_working_day definition with valid params")
    @pytest.mark.caseid("C46247467")
    @pytest.mark.parametrize(
        "date, calendars, currencies, holiday_outputs",
        [
            (
                "2018-12-31",
                ["UKR", "FRA"],
                ["EUR"],
                [HolidayOutputs.DATE, HolidayOutputs.NAMES, HolidayOutputs.COUNTRIES],
            ),
            (
                timedelta(-60),
                None,
                ["EUR"],
                ["Date", HolidayOutputs.NAMES, HolidayOutputs.COUNTRIES],
            ),
            (datetime(2022, 4, 13), ["UKR", "FRA"], None, ["Date", "Names"]),
        ],
    )
    def test_is_working_day_definition_with_valid_params(
        self,
        open_session,
        date,
        calendars,
        currencies,
        holiday_outputs,
    ):
        response = is_working_day.Definition(
            date=date,
            calendars=calendars,
            currencies=currencies,
            holiday_outputs=holiday_outputs,
        ).get_data()

        assert isinstance(response.data.day, WorkingDay)
        assert isinstance(response.data.day.holidays, list)
        assert isinstance(response.data.day.is_working_day, bool)

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(
            response, ["isWorkingDay", "isWeekEnd"]
        )

    @allure.title(
        "Create a is_working_day definition with valid and invalid definition asynchronously"
    )
    @pytest.mark.caseid("C46670891")
    async def test_is_working_day_definition_with_valid_and_invalid_def_async(
        self, open_session_async
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            first_definition, invalid_definition
        )

        check_non_empty_response_data(valid_response)
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="IPA_library.ErrorBusinessCalendar_CalendarNotFound",
            expected_error_message="Business Calendar: calendar not found.",
        )

    @allure.title(
        "Create a is_working_day definition with invalid calendar and getting RDError"
    )
    @pytest.mark.parametrize(
        "expected_error",
        [
            (
                "Error code IPA_library.ErrorBusinessCalendar_CalendarNotFound | Business Calendar: calendar not found."
            )
        ],
    )
    @pytest.mark.caseid("C46247474")
    def test_is_working_day_definition_with_invalid_param(
        self, open_session, expected_error
    ):
        with pytest.raises(RDError) as error:
            is_working_day.Definition(
                date=datetime.now(),
                currencies=["UKR"],
            ).get_data()
        assert str(error.value) == expected_error

    @allure.title("Create is_working_day definition with extended params and get data")
    @pytest.mark.parametrize(
        "extended_param",
        [
            {
                "date": "2022-09-22",
                "calendars": ["USA"],
            }
        ],
    )
    @pytest.mark.caseid("C46247470")
    def test_is_working_day_definition_with_extended_params(
        self, open_session, extended_param
    ):
        response = is_working_day.Definition(
            tag="my request",
            date=timedelta(-360),
            calendars=["UKR", "FRA"],
            currencies=["EUR"],
            extended_params=extended_param,
        ).get_data()

        check_extended_params_were_sent_in_request(response, extended_param)
        check_non_empty_response_data(response)
        check_response_status(
            response=response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

        assert response.data.day.is_working_day is True
        assert response.data.day.is_weekend is False

    @allure.title(
        "Create multi is_working_day definitions for working days and holidays and get data"
    )
    @pytest.mark.caseid("C46247473")
    def test_is_working_day_definitions_with_valid_def(self, open_session):
        response = is_working_day.Definitions(
            [first_definition, second_definition]
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(
            response, ["isWorkingDay", "isWeekEnd"]
        )

        assert isinstance(response.data.days, list)
        assert isinstance(response.data.days[0].holidays, list)
        assert response.data.days[0].is_working_day is False



# coding: utf8

from typing import Optional

from ..._instrument_definition import ObjectDefinition


class FxForwardStart(ObjectDefinition):
    """
    Parameters
    ----------
    forward_start_date : str, optional
        Expiry date of the Forward Start option
    forward_start_tenor : str, optional
        Tenor of the Forward Start option
    strike_percent : float, optional
        Strike Percent of the Forward Start date of the option
    """

    def __init__(
        self,
        forward_start_date: Optional[str] = None,
        forward_start_tenor: Optional[str] = None,
        strike_percent: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.forward_start_date = forward_start_date
        self.forward_start_tenor = forward_start_tenor
        self.strike_percent = strike_percent

    @property
    def forward_start_date(self):
        """
        Expiry date of the Forward Start option
        :return: str
        """
        return self._get_parameter("forwardStartDate")

    @forward_start_date.setter
    def forward_start_date(self, value):
        self._set_parameter("forwardStartDate", value)

    @property
    def forward_start_tenor(self):
        """
        Tenor of the Forward Start option
        :return: str
        """
        return self._get_parameter("forwardStartTenor")

    @forward_start_tenor.setter
    def forward_start_tenor(self, value):
        self._set_parameter("forwardStartTenor", value)

    @property
    def strike_percent(self):
        """
        Strike Percent of the Forward Start date of the option
        :return: float
        """
        return self._get_parameter("strikePercent")

    @strike_percent.setter
    def strike_percent(self, value):
        self._set_parameter("strikePercent", value)



# coding: utf8
from typing import Optional

from ._swap_leg_definition import LegDefinition
from .._instrument_definition import InstrumentDefinition


class SwapInstrumentDefinition(InstrumentDefinition):
    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        instrument_code: Optional[str] = None,
        trade_date: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        legs: Optional[LegDefinition] = None,
        is_non_deliverable: Optional[bool] = None,
        settlement_ccy: Optional[str] = None,
        start_tenor: Optional[str] = None,
        template: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.instrument_tag = instrument_tag
        self.instrument_code = instrument_code
        self.trade_date = trade_date
        self.start_date = start_date
        self.end_date = end_date
        self.tenor = tenor
        self.legs = legs
        self.is_non_deliverable = is_non_deliverable
        self.settlement_ccy = settlement_ccy
        self.start_tenor = start_tenor
        self.template = template

    @classmethod
    def get_instrument_type(cls):
        return "Swap"

    @property
    def legs(self):
        """
        The legs of the Swap to provide a full definition of the swap if no template or instrumentCode have been provided.
        Optional. Either InstrumentCode, Template, or Legs must be provided.
        :return: list SwapLegDefinition
        """
        return self._get_list_parameter(LegDefinition, "legs")

    @legs.setter
    def legs(self, value):
        self._set_list_parameter(LegDefinition, "legs", value)

    @property
    def end_date(self):
        """
        The maturity date of the swap contract.
        Mandatory. Either the endDate or the tenor must be provided.
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def instrument_code(self):
        """
        A swap RIC that is used to retrieve the description of the swap contract.
        Optional. Either instrumentCode, template, or legs must be provided.
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)

    @property
    def is_non_deliverable(self):
        """
        A flag that indicates if the swap is non-deliverable.
        Optional. By defaults 'false'.
        :return: bool
        """
        return self._get_parameter("isNonDeliverable")

    @is_non_deliverable.setter
    def is_non_deliverable(self, value):
        self._set_parameter("isNonDeliverable", value)

    @property
    def settlement_ccy(self):
        """
        For non-deliverable instrument, the ISO code of the settlement currency.
        Optional. By priority order : 'USD' if one leg denominated in USD; 'EUR' if one leg is denominated in EUR; the paidLegCcy.
        :return: str
        """
        return self._get_parameter("settlementCcy")

    @settlement_ccy.setter
    def settlement_ccy(self, value):
        self._set_parameter("settlementCcy", value)

    @property
    def start_date(self):
        """
        The date the swap starts accruing interest. Its effective date.
        Optional. By default, it is derived from the TradeDate and the day to spot convention of the contract currency.
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        self._set_parameter("startDate", value)

    @property
    def start_tenor(self):
        """
        The code indicating the period from a spot date to startdate of the instrument
        (e.g. '1m'). no default value applies.
        :return: str
        """
        return self._get_parameter("startTenor")

    @start_tenor.setter
    def start_tenor(self, value):
        self._set_parameter("startTenor", value)

    @property
    def template(self):
        """
        A reference to a common swap contract.
        Optional. Either InstrumentCode, Template, or Legs must be provided.
        :return: str
        """
        return self._get_parameter("template")

    @template.setter
    def template(self, value):
        self._set_parameter("template", value)

    @property
    def tenor(self):
        """
        The period code that represents the time between the start date and end date the contract.
        Mandatory. Either the endDate or the tenor must be provided.
        :return: str
        """
        return self._get_parameter("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_parameter("tenor", value)

    @property
    def trade_date(self):
        """
        The date the swap contract was created.
        Optional. By default, the valuation date.
        :return: str
        """
        return self._get_parameter("tradeDate")

    @trade_date.setter
    def trade_date(self, value):
        self._set_parameter("tradeDate", value)



from .context_collection import RDPMixin, UDFMixin
from ._history_provider import HistoryProvider
from ..content.fundamental_and_reference._data_grid_type import (
    DataGridType,
)


class HistoryRDPProvider(RDPMixin, HistoryProvider):
    pass


class HistoryUDFProvider(UDFMixin, HistoryProvider):
    pass


provider_by_data_grid_type = {
    DataGridType.UDF: HistoryUDFProvider(),
    DataGridType.RDP: HistoryRDPProvider(),
}


def get_history_provider(
    data_grid_type: "DataGridType",
) -> "HistoryProvider":
    provider = provider_by_data_grid_type.get(data_grid_type)

    if not provider:
        raise TypeError(f"Unexpected platform type. Type: {data_grid_type}")

    return provider



import threading

import pytest
from mock.mock import patch

from refinitiv.data.delivery._stream import RDPStreamConnection
from refinitiv.data.delivery._stream.stream_cxn_state import StreamCxnState
from tests.unit.conftest import StubSession
from tests.unit.delivery.stream.conftest import StubStreamCxnConfig, StubWebSocketApp
import sys


# --------------------------------------------------------------------------------------
#
# TEST TRANSITIONS
#
# --------------------------------------------------------------------------------------


def test_after_cxn_is_created_state_is_initial():
    # given
    sess = StubSession()
    cfg = StubStreamCxnConfig()

    # when
    cxn = RDPStreamConnection(connection_id=1, name="name", session=sess, config=cfg)

    # then
    assert cxn.state is StreamCxnState.Initial, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_initial_to_connecting(create_omm_cxn):
    # given
    cxn = create_omm_cxn()

    assert cxn.state is StreamCxnState.Initial, cxn.state

    # when
    cxn.wait_start()

    # then
    assert cxn.is_connecting is True, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_connecting_to_message_processing(create_omm_cxn):
    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws

    # when
    ws.cmd("on_open")
    # here ws sends login message
    assert cxn.is_connecting is True, cxn.state

    # response
    message = '[{"State": {"Stream": "Open"}}]'
    ws.cmd("on_message", message)

    # then
    assert cxn.is_message_processing is True, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_message_processing_to_disconnecting(create_omm_cxn):
    def get_response():
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)

    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws
    ws.cmd("on_open")
    # here ws sends login message

    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=get_response, daemon=True).start()
    cxn.wait_connection_result()

    assert cxn.is_message_processing is True, cxn.state

    cxn.start_disconnecting()

    assert cxn.is_disconnecting, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_disconnecting_to_disconnected(create_omm_cxn):
    def get_response():
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)

    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws
    ws.cmd("on_open")
    # here ws sends login message
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=get_response, daemon=True).start()
    cxn.wait_connection_result()
    cxn.start_disconnecting()

    assert cxn.is_disconnecting, cxn.state

    cxn.end_disconnecting()

    assert cxn.is_disconnected, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_disconnected_to_disposed(create_omm_cxn):
    def get_response():
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)

    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws
    ws.cmd("on_open")
    # here ws sends login message
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=get_response, daemon=True).start()
    cxn.wait_connection_result()
    cxn.start_disconnecting()
    cxn.end_disconnecting()

    assert cxn.is_disconnected, cxn.state

    cxn.dispose()

    assert cxn.is_disposed, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_connecting_to_disconnected(create_omm_cxn):
    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws

    # when
    ws.cmd("on_open")
    # here ws sends login message

    assert cxn.is_connecting is True, cxn.state

    # response
    message = '[{"State": {"Stream": "Close"}}]'
    ws.cmd("on_message", message)
    ws.cmd("on_close")
    cxn.wait_connection_result()

    # then
    assert cxn.is_disposed is True, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_connecting_to_disposed(create_omm_cxn):
    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws

    # when
    ws.cmd("on_open")
    # here ws sends login message

    assert cxn.is_connecting is True, cxn.state

    cxn.dispose()

    # then
    assert cxn.is_disposed is True, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_connecting_cannot_to_disconnecting(create_omm_cxn):
    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws

    # when
    ws.cmd("on_open")
    # here ws sends login message

    assert cxn.is_connecting is True, cxn.state

    cxn.start_disconnecting()

    assert cxn.is_connecting is True, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_message_processing_cannot_to_disposed(create_omm_cxn):
    def get_response():
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)

    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws
    ws.cmd("on_open")
    # here ws sends login message
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=get_response, daemon=True).start()
    cxn.wait_connection_result()

    assert cxn.is_message_processing is True, cxn.state

    cxn.dispose()

    assert cxn.is_message_processing is True, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_message_processing_cannot_to_disconnected(create_omm_cxn):
    def get_response():
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)

    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws
    ws.cmd("on_open")
    # here ws sends login message
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=get_response, daemon=True).start()
    cxn.wait_connection_result()

    assert cxn.is_message_processing is True, cxn.state

    cxn.end_disconnecting()

    assert cxn.is_message_processing is True, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_from_message_processing_cannot_to_connecting_if_not_auto_reconnect(
    create_omm_cxn,
):
    def get_response():
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)

    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws
    ws.cmd("on_open")
    # here ws sends login message
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=get_response, daemon=True).start()
    cxn.wait_connection_result()

    assert cxn.is_message_processing is True, cxn.state

    ws.cmd("on_close")
    cxn.join()

    assert cxn.is_disposed is True, cxn.state


# --------------------------------------------------------------------------------------


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_cxn_cannot_cycle_infos_if_not_auto_reconnect(create_omm_cxn):
    def try_cycle_infos():
        message = '[{"State": {"Text": "Login Rejected."}}]'
        ws.cmd("on_message", message)
        ws.cmd("on_close")

    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws
    ws.cmd("on_open")

    # here ws sends login message

    def next_available_info_mock():
        assert False

    cxn._config.next_available_info = next_available_info_mock
    prev_info = cxn._config.info
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=try_cycle_infos, daemon=True).start()
    cxn.wait_connection_result()
    curr_info = cxn._config.info

    assert prev_info == curr_info
    assert cxn.is_disconnected, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_if_call_dispose_when_wait_connection_result_it_will_disposed(create_omm_cxn):
    def call_dispose():
        cxn.dispose()

    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws
    ws.cmd("on_open")
    # here ws sends login message
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=call_dispose, daemon=True).start()
    cxn.wait_connection_result()

    # then
    assert cxn.is_disposed, cxn.state


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_when_connection_result_is_success_state_is_message_processing(create_omm_cxn):
    def open_message():
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)

    # given
    cxn = create_omm_cxn()
    cxn.wait_start()
    ws = cxn._listener.ws
    ws.cmd("on_open")
    # here ws sends login message
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=open_message, daemon=True).start()
    cxn.wait_connection_result()

    # then
    assert cxn.is_message_processing, cxn.state



from refinitiv.data.content.custom_instruments import Definition
from refinitiv.data.delivery._data._request import Request
from tests.unit.conftest import StubSession, StubResponse

get_uuid_url = "test_get_rdp_url_root/data/custom-instruments/v1/instruments/S%29Instrument.UUID-0000"


def test_get_uui_request_only_once():
    request_counter = 0
    my_uuid = "MY-0000"

    def _http_request(request: Request):
        nonlocal request_counter
        if request.url == get_uuid_url:
            request_counter += 1
            return StubResponse(
                content_data={
                    "state": {
                        "code": 400,
                        "status": "Bad Request",
                        "message": "Validation Error",
                    },
                    "data": [
                        {
                            "key": "symbol",
                            "reason": f".UUID suffix UUID-0000 not matched with userID {my_uuid}",
                        }
                    ],
                },
                status_code=400,
            )

    session = StubSession(is_open=True)
    session.http_request = _http_request

    universe = ["Instrument1.AA-0000", f"S)Instrument2.{my_uuid}"]
    stream = Definition(universe=universe).get_stream(session=session)
    _ = stream["INST"]
    _ = stream["Instrument2"]
    _ = stream["Instrument1"]
    _ = stream[universe[0]]
    assert request_counter == 1


def test_all_universe_with_uuid():
    my_uuid = "MY-0000"

    def _http_request(request: Request):
        if request.url == get_uuid_url:
            return StubResponse(
                content_data={
                    "state": {
                        "code": 400,
                        "status": "Bad Request",
                        "message": "Validation Error",
                    },
                    "data": [
                        {
                            "key": "symbol",
                            "reason": f".UUID suffix UUID-0000 not matched with userID {my_uuid}",
                        }
                    ],
                },
                status_code=400,
            )

    session = StubSession(is_open=True)
    session.http_request = _http_request

    universe = ["Instrument1.AA-0000", f"S)Instrument2.{my_uuid}"]
    stream = Definition(universe=universe).get_stream(session=session)

    result = set(stream_item.name for stream_item in stream)
    expected_result = {"S)Instrument1.AA-0000", f"S)Instrument2.{my_uuid}"}
    assert result == expected_result
    assert stream["Instrument1.AA-0000"].name == "S)Instrument1.AA-0000"
    assert stream[f"Instrument2.{my_uuid}"].name == universe[1]
    assert stream["Instrument2"].name == universe[1]
    assert stream["Instrument1"] == {}


def test_mixed_universe():
    my_uuid = "MY-0000"

    def _http_request(request: Request):
        if request.url == get_uuid_url:
            return StubResponse(
                content_data={
                    "state": {
                        "code": 400,
                        "status": "Bad Request",
                        "message": "Validation Error",
                    },
                    "data": [
                        {
                            "key": "symbol",
                            "reason": f".UUID suffix UUID-0000 not matched with userID {my_uuid}",
                        }
                    ],
                },
                status_code=400,
            )

    session = StubSession(is_open=True)
    session.http_request = _http_request

    universe = [
        "S)Instrument1.AA-0000",
        f"S)Instrument2.{my_uuid}",
        "Instrument1",
        "Instrument3",
    ]
    stream = Definition(universe=universe).get_stream(session=session)

    result = set(stream_item.name for stream_item in stream)
    expected_result = {
        "S)Instrument1.AA-0000",
        f"S)Instrument2.{my_uuid}",
        f"S)Instrument1.{my_uuid}",
        f"S)Instrument3.{my_uuid}",
    }

    assert result == expected_result
    assert stream["Instrument1.AA-0000"].name == "S)Instrument1.AA-0000"
    assert stream[f"Instrument2.{my_uuid}"].name == universe[1]
    assert stream["Instrument2"].name == universe[1]
    assert stream["Instrument1"].name == f"S)Instrument1.{my_uuid}"


def test_add_instruments():
    # given
    test_universe = ["Instrument1", "Instrument2"]
    test_items = ["Instrument3"]
    test_result = ["S)Instrument1.", "S)Instrument2.", "S)Instrument3."]
    session = StubSession(is_open=True)
    stream = Definition(universe=test_universe).get_stream(session=session)

    # when
    stream.add_instruments(test_items)

    # then
    assert stream._universe == test_result


def test_remove_instruments():
    # given
    test_universe = ["Instrument1", "Instrument2"]
    test_items = ["Instrument2"]
    test_result = ["S)Instrument1."]
    session = StubSession(is_open=True)
    stream = Definition(universe=test_universe).get_stream(session=session)

    # when
    stream.remove_instruments(test_items)

    # then
    assert stream._universe == test_result


def test_add_fields():
    # given
    test_fields = ["BID", "ASK"]
    test_new_fields = ["TRDPRC_1"]
    test_result = ["BID", "ASK", "TRDPRC_1"]
    session = StubSession(is_open=True)
    stream = Definition(universe=["VOD.L"], fields=test_fields).get_stream(
        session=session
    )

    # when
    stream.add_fields(test_new_fields)

    # then
    assert stream._fields == test_result


def test_remove_fields():
    # given
    test_fields = ["BID", "ASK"]
    test_remove_fields = ["ASK"]
    test_result = ["BID"]
    session = StubSession(is_open=True)
    stream = Definition(universe=["VOD.L"], fields=test_fields).get_stream(
        session=session
    )

    # when
    stream.remove_fields(test_remove_fields)

    # then
    assert stream._fields == test_result



# coding: utf8

__all__ = ["StrikeFilter"]

from ..._object_definition import ObjectDefinition


class StrikeFilter(ObjectDefinition):
    def __init__(self, max_of_median_implied_vol=None, min_of_median_implied_vol=None):
        super().__init__()
        self.max_of_median_implied_vol = max_of_median_implied_vol
        self.min_of_median_implied_vol = min_of_median_implied_vol

    @property
    def max_of_median_implied_vol(self):
        """
        Remove strikes whose implied vol is more than MaxOfMedianImpliedVolPercent x Median implied Vol.
        :return: float
        """
        return self._get_parameter("maxOfMedianImpliedVol")

    @max_of_median_implied_vol.setter
    def max_of_median_implied_vol(self, value):
        self._set_parameter("maxOfMedianImpliedVol", value)

    @property
    def min_of_median_implied_vol(self):
        """
        Remove strikes whose implied vol is less than MinOfMedianImpliedVolPercent x Median implied Vol.
        :return: float
        """
        return self._get_parameter("minOfMedianImpliedVol")

    @min_of_median_implied_vol.setter
    def min_of_median_implied_vol(self, value):
        self._set_parameter("minOfMedianImpliedVol", value)



"""Define basic abstract classes for templates functionality

That is abstracted from the original goal - discovery search templates.
And can be used for any other Definition or any other callable.
"""

from typing import Optional, Dict, Any, Set, Iterable

from refinitiv.data._tools.templates import StringTemplate

from .namespaces import Namespace


class Target:
    """Abstract callable target with accessible list of keyword arguments names

    Base for building templates in TargetTemplate.
    You need to define your own targets that do some work. Like DiscoverySearchTarget.
    """

    def __init__(self):
        self.args_names: Set[str] = set()

    def __call__(self, **kwargs):
        pass


def extract_used_templates_from_placeholders_and_namespace(
    names: Set[str], namespace: Namespace
) -> Set[str]:
    """Get a full path to namespaces that was used in the set of template placeholders

    In the string template we can have placeholders like
    "builtins.utils.geo.Mines.location.lat". And it's impossible to decide just by
    looking on it what part of this string is namespaces chain (prefix), what is
    Template name, and what is chain of attributes of the Template call result (suffix).
    The goal of this functions is to find maximum namespaces + Template name chain
    based on given namespace.

    For example, if in namespace we have template with the namespace chain
    "builtings.utils.geo" and name "Mines", we will return "builtins.utils.geo.Mines".
    And what left of original string must be processed as a result attributes later.

    You must be sure that you are passing as a names arguments all names
    that MUST BE treated as a subtemplate usage. In other case, exception will be raised.

    Parameters
    ----------

    names: Set[str]
        set of placeholder names, without suffixes like jinja filters, but with
        attributes of the expected subtemplate output
    namespace: Namespace
        namespace object of your template

    Returns
    -------
    Set[str]
        Templates used, full paths, without attributes

    Raises
    ------
    ValueError
        If it's impossible to find template for one of your placeholders
    """

    # pretending we can't have template on the root on namespace here
    ns_prefixes = namespace.keys()

    def match_name(orig_name):
        """Check if orig_name or its prefix is presented in the namespace

        Function expects that orig_name already  starts with one of the root namespace
        prefixes. So it always matches at least to this prefix.
        """
        cur_name = orig_name

        while True:
            try:
                found = namespace.get(cur_name)
            except KeyError:
                raise ValueError(f"Unknown subtemplate usage: {orig_name}")

            if found:
                break
            else:
                # cut the tail
                parts = cur_name.rsplit(".", maxsplit=1)
                cur_name = parts[0]

        return cur_name

    result = set()

    for name in names:
        prefix = name.split(".", maxsplit=1)[0]

        if prefix in ns_prefixes:
            matched = match_name(name)
            if matched and matched not in ns_prefixes:
                result.add(matched)

    return result


class TargetTemplate:
    """Abstract target search preset

    Initialized with default values for defined Target
    Any string value acts as template string. You can use placeholder variables,
    and those variables will be required to prepare search parameters through
    `._search_kwargs()` or to launch search through `.search()`.

    Placeholder variables syntax based on jinja2 with some changes:
    - Variables defined with #{varname} syntax
    - Instructions with {{}} symbols. Available instructions:
      - {{if}}<one value>{{else}}<other value>{{endif}}

    Attributes
    ----------

    name: str
        name of the template

    """

    _target_class = Target

    def __init__(
        self,
        name=None,
        *,
        placeholders_defaults: Optional[Dict[str, Any]] = None,
        pass_through_defaults: Optional[Dict[str, Any]] = None,
        optional_placeholders: Iterable[str] = None,
        ns: "Namespace" = None,
        **search_defaults,
    ):
        """
        Parameters
        ----------
        name : str, optional
            name of the template
        placeholders_defaults: dict, optional
            Dict of string template placeholders default values.
        pass_through_defaults: dict, optional
            default values for the Target parameters
        optional_placeholders: Iterable[str], optional
            names of placeholders that are optional without default values
        ns: Namespace
            Namespace in which template will operate. Used for subtemplates.
        """

        self._ns = ns if ns is not None else Namespace()
        self._target: Target = self._target_class()
        # Names of the templates used, with full path from self.ns root
        self._subtemplates_used = set()

        """ List search keyword arguments we can use in this template """
        self._placeholders_defaults = (
            {} if placeholders_defaults is None else placeholders_defaults
        )
        self._optional_placeholders = set(
            [] if optional_placeholders is None else optional_placeholders
        )
        """ Default template variables values for a templated defaults """
        if pass_through_defaults is None:
            pass_through_defaults = {}

        bad_pass_through_params = set(pass_through_defaults) - self._target.args_names
        if bad_pass_through_params:
            raise ValueError(
                "All the parameters described in 'parameters' section of search "
                "template configuration, must be either placeholders variables or "
                "parameters of the discovery search Definition. These parameters are "
                "neither of them: " + ", ".join(bad_pass_through_params)
            )

        self.name = name

        unknown_defaults = set(search_defaults) - self._target.args_names
        if unknown_defaults:
            raise ValueError(
                "These arguments are defined in template, but not in search Definition: "
                + ", ".join(unknown_defaults)
            )
        # Names of all placeholders inside string templates
        """Set of names for all placeholders in string templates"""
        self._placeholder_names: Set[str] = set()
        # Arguments to be passed to Definition as templates
        self._templated_defaults: Dict[str, StringTemplate] = {}
        # Arguments to be directly passed to Definition without any preprocessing
        self._pass_through_defaults: Dict[str, Any] = {}

        def is_subtemplate(placeholder_name: str):
            return any(
                placeholder_name.startswith(prefix) for prefix in self._ns.keys()
            )

        for name, value in search_defaults.items():

            if not isinstance(value, str):
                self._pass_through_defaults[name] = value
                continue

            template = StringTemplate(value)
            template.validate()
            if template.placeholders():
                self._templated_defaults[name] = template
                for name in template.placeholders():
                    if not is_subtemplate(name):
                        self._placeholder_names.add(name)
            else:
                self._pass_through_defaults[name] = value

            self._subtemplates_used |= (
                extract_used_templates_from_placeholders_and_namespace(
                    template.placeholders(), self._ns
                )
            )

        self._pass_through_defaults.update(pass_through_defaults)

        bad_tpl_var_names = self._get_full_placeholder_names() & self._target.args_names
        if bad_tpl_var_names:
            raise ValueError(
                "You can't use template arguments with the same name"
                " as search arguments. You have used: " + ", ".join(bad_tpl_var_names)
            )

    def _get_full_placeholder_names(self):
        placeholder_names = set()
        for st_path in self._subtemplates_used:
            placeholder_names.update(
                self._ns.get(st_path)._get_full_placeholder_names()
            )
        return self._placeholder_names | placeholder_names

    def __repr__(self):
        return f"<TargetTemplate for {self._target_class.__name__} name='{self.name}'>"

    def _search_kwargs(self, **kwargs) -> dict:
        """Get dictionary of arguments for the Target"""

        undefined_placeholders = (
            self._get_full_placeholder_names()
            - set(kwargs)
            - set(self._placeholders_defaults)
            - self._optional_placeholders
        )

        assert all("." not in ph for ph in undefined_placeholders)

        if undefined_placeholders:
            raise KeyError(
                "Those keyword arguments must be defined, but they are not: "
                + ", ".join(undefined_placeholders)
            )

        unexpected_arguments = (
            set(kwargs)
            - self._get_full_placeholder_names()
            # templated defaults can't be redefined
            - (self._target.args_names - self._templated_defaults.keys())
        )

        if unexpected_arguments:
            raise KeyError(f"Unexpected arguments: {', '.join(unexpected_arguments)}")

        old_kwargs = kwargs
        kwargs = kwargs.copy()
        # Applying template variables defaults
        for name, value in self._placeholders_defaults.items():
            if name not in kwargs:
                kwargs[name] = value

        def set_path_value(path, value):
            """Set value to kwargs dict of dicts by path with dots

            Expecting to always have at least one dot in path, because subtemplate
            always has a prefix.
            """
            parts = path.split(".")
            assert len(parts) > 1
            current = kwargs
            for part in parts[:-1]:
                if part not in current:
                    current[part] = {}
                current = current[part]
            current[parts[-1]] = value

        for subname in self._subtemplates_used:
            st = self._ns.get(subname)
            filtered_kwargs = {
                k: v
                for k, v in old_kwargs.items()
                if k in st._get_full_placeholder_names()
            }
            set_path_value(subname, st.search(**filtered_kwargs))

        result = self._pass_through_defaults.copy()

        # Apply variables to templated defaults
        for name, template in self._templated_defaults.items():
            result[name] = template.substitute(**kwargs)

        # Apply other variables from kwargs
        for name, value in kwargs.items():
            if (
                name not in self._get_full_placeholder_names()
                and name not in self._ns.keys()
            ):
                result[name] = value

        return result

    def search(self, **kwargs):
        """Target call with given parameters"""
        return self._target(**self._search_kwargs(**kwargs))



from typing import Any, Callable, TYPE_CHECKING, Optional

from ._rdp_stream import _RDPStream
from ._stream_factory import create_rdp_stream
from .base_stream import StreamOpenMixin
from ..._content_type import ContentType
from ..._core.session import get_valid_session
from ..._tools import cached_property, create_repr, make_callback

if TYPE_CHECKING:
    from ..._types import ExtendedParams
    from ... import OpenState
    from ..._core.session import Session


class RDPStream(StreamOpenMixin):
    """
    Open an RDP stream.

    Parameters
    ----------

    service: string, optional
        name of RDP service

    universe: list
        RIC to retrieve item stream.

    view: list
        data fields to retrieve item stream

    parameters: dict
        extra parameters to retrieve item stream.

    api: string
        specific name of RDP streaming defined in config file. i.e.
        'streaming/trading-analytics/redi'

    extended_params: dict, optional
        Specify optional params
        Default: None

    Raises
    ------
    Exception
        If request fails or if Refinitiv Services return an error

    Examples
    --------
    >>> import refinitiv.data as rd
    >>> APP_KEY = "APP_KEY"
    >>> USERNAME = "USERNAME"
    >>> PASSWORD = "PASSWORD"
    >>> session = rd.session.platform.Definition(
    ...         app_key=APP_KEY,
    ...         grant=rd.session.platform.GrantPassword(
    ...             username=USERNAME,
    ...             password=PASSWORD,
    ...         )
    ... ).get_session()
    >>> session.open()
    >>>
    >>> tds = rd.delivery.rdp_stream.Definition(
    ...     service="",
    ...     universe=[],
    ...     view=[],
    ...     parameters={"universeType": "RIC"},
    ...     api='streaming.trading-analytics.endpoints.redi'
    ... ).get_stream(session)
    >>> tds.open()
    """

    def __init__(
        self,
        service: str,
        universe: list,
        view: list,
        parameters: dict,
        api: str,
        session: Optional["Session"] = None,
        extended_params: "ExtendedParams" = None,
    ) -> None:
        self._session = get_valid_session(session)
        self._always_use_default_session = session is None
        self._service = service
        self._universe = universe
        self._view = view
        self._parameters = parameters
        self._api = api
        self._extended_params = extended_params

    @cached_property
    def _stream(self) -> _RDPStream:
        return create_rdp_stream(
            ContentType.STREAMING_CUSTOM,
            api=self._api,
            session=self._session,
            service=self._service,
            universe=self._universe,
            view=self._view,
            parameters=self._parameters,
            extended_params=self._extended_params,
        )

    def open(self) -> "OpenState":
        """
        Opens the RDPStream to start to stream. Once it's opened,
        it can be used in order to retrieve data.

        Parameters
        ----------

        Returns
        -------
        OpenState
            current state of this RDP stream object.

        Examples
        --------
        >>> from refinitiv.data.delivery import rdp_stream
        >>> definition = rdp_stream.Definition(
        ...                    service=None,
        ...                    universe=[],
        ...                    view=None,
        ...                    parameters={"universeType": "RIC"},
        ...                    api='streaming.trading-analytics.endpoints.redi')
        >>> stream = definition.get_stream()
        >>> await stream.open_async()
        """
        return super().open()

    async def open_async(self) -> "OpenState":
        """
        Opens asynchronously the RDPStream to start to stream

        Parameters
        ----------

        Returns
        -------
        OpenState
            current state of this RDP stream object.

        Examples
        --------
        >>> from refinitiv.data.delivery import rdp_stream
        >>> definition = rdp_stream.Definition(
        ...                     service=None,
        ...                     universe=[],
        ...                     view=None,
        ...                     parameters={"universeType": "RIC"},
        ...                     api='streaming.trading-analytics.endpoints.redi')
        >>> stream = definition.get_stream()
        >>> await stream.open_async()
        """
        return await super().open_async()

    def close(self) -> "OpenState":
        """
        Closes the RPDStream connection, releases resources

        Returns
        -------
        OpenState
            current state of this RDP stream object.

        Examples
        --------
        >>> from refinitiv.data.delivery import rdp_stream
        >>> definition = rdp_stream.Definition(
        ...                    service=None,
        ...                    universe=[],
        ...                    view=None,
        ...                    parameters={"universeType": "RIC"},
        ...                    api='streaming.trading-analytics.endpoints.redi')
        >>> stream = definition.get_stream()
        >>> stream.open()
        >>> stream.close()
        """
        return super().close()

    def on_ack(self, on_ack: Callable[[dict, "RDPStream"], Any]) -> "RDPStream":
        """
        This function called when the stream received an ack message.

        Parameters
        ----------
        on_ack : Callable, optional
             Callable object to process retrieved ack data

        Returns
        -------
        RDPStream
            current instance it is a RDP stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import rdp_stream
        >>>
        >>> def display_response(event, stream):
        ...     print(f'Ack received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = rdp_stream.Definition(
        ...                     service=None,
        ...                     universe=[],
        ...                     view=None,
        ...                     parameters={"universeType": "RIC"},
        ...                     api='streaming.trading-analytics.endpoints.redi')
        >>> stream = definition.get_stream()
        >>> stream.on_ack(display_response)
        >>>
        >>> stream.open()
        """
        self._stream.on_ack(make_callback(on_ack))
        return self

    def on_response(
        self, on_response: Callable[[dict, "RDPStream"], Any]
    ) -> "RDPStream":
        """
        This function called when the stream received an response message.

        Parameters
        ----------
        on_response : Callable, optional
             Callable object to process retrieved response data

        Returns
        -------
        RDPStream
            current instance it is a RDP stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import rdp_stream
        >>>
        >>> def display_response(event, stream):
        ...     print(f'Response received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = rdp_stream.Definition(
        ...                    service=None,
        ...                    universe=[],
        ...                    view=None,
        ...                    parameters={"universeType": "RIC"},
        ...                    api='streaming.trading-analytics.endpoints.redi')
        >>> stream = definition.get_stream()
        >>> stream.on_response(display_response)
        >>>
        >>> stream.open()
        """
        self._stream.on_response(make_callback(on_response))
        return self

    def on_update(self, on_update: Callable[[dict, "RDPStream"], Any]) -> "RDPStream":
        """
        This function called when the stream received an update message.

        Parameters
        ----------
        on_update : Callable, optional
             Callable object to process retrieved update data

        Returns
        -------
        RDPStream
            current instance it is a RDP stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import rdp_stream
        >>>
        >>> def display_response(event, stream):
        ...     print(f'Update received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = rdp_stream.Definition(
        ...                    service=None,
        ...                    universe=[],
        ...                    view=None,
        ...                    parameters={"universeType": "RIC"},
        ...                    api='streaming.trading-analytics.endpoints.redi')
        >>> stream = definition.get_stream()
        >>> stream.on_update(display_response)
        >>>
        >>> stream.open()
        """
        self._stream.on_update(make_callback(on_update))
        return self

    def on_alarm(self, on_alarm: Callable[[dict, "RDPStream"], Any]) -> "RDPStream":
        """
        This function called when the stream received an alarm message.

        Parameters
        ----------
        on_alarm : Callable, optional
             Callable object to process retrieved alarm data

        Returns
        -------
        RDPStream
            current instance it is a RDP stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import rdp_stream
        >>>
        >>> def display_response(event, stream):
        ...     print(f'Alarm received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = rdp_stream.Definition(
        ...                    service=None,
        ...                    universe=[],
        ...                    view=None,
        ...                    parameters={"universeType": "RIC"},
        ...                    api='streaming.trading-analytics.endpoints.redi')
        >>> stream = definition.get_stream()
        >>> stream.on_alarm(display_response)
        >>>
        >>> stream.open()
        """
        self._stream.on_alarm(make_callback(on_alarm))
        return self

    def __repr__(self):
        return create_repr(
            self,
            middle_path="rdp_stream",
            class_name=self.__class__.__name__,
        )



# coding: utf8

from enum import Enum, unique


@unique
class RedemptionDateType(Enum):
    REDEMPTION_AT_AVERAGE_LIFE = "RedemptionAtAverageLife"
    REDEMPTION_AT_BEST_DATE = "RedemptionAtBestDate"
    REDEMPTION_AT_CALL_DATE = "RedemptionAtCallDate"
    REDEMPTION_AT_CUSTOM_DATE = "RedemptionAtCustomDate"
    REDEMPTION_AT_MAKE_WHOLE_CALL_DATE = "RedemptionAtMakeWholeCallDate"
    REDEMPTION_AT_MATURITY_DATE = "RedemptionAtMaturityDate"
    REDEMPTION_AT_NEXT_DATE = "RedemptionAtNextDate"
    REDEMPTION_AT_PAR_DATE = "RedemptionAtParDate"
    REDEMPTION_AT_PARTIAL_CALL_DATE = "RedemptionAtPartialCallDate"
    REDEMPTION_AT_PARTIAL_PUT_DATE = "RedemptionAtPartialPutDate"
    REDEMPTION_AT_PERPETUITY = "RedemptionAtPerpetuity"
    REDEMPTION_AT_PREMIUM_DATE = "RedemptionAtPremiumDate"
    REDEMPTION_AT_PUT_DATE = "RedemptionAtPutDate"
    REDEMPTION_AT_SINK_DATE = "RedemptionAtSinkDate"
    REDEMPTION_AT_WORST_DATE = "RedemptionAtWorstDate"



from typing import List, Optional, TYPE_CHECKING, Union

from numpy import iterable

from ..._curves._interest_rate_curve_get_definition import (
    InterestRateCurveGetDefinition,
)
from ...._content_provider_layer import ContentProviderLayer
from ....._content_type import ContentType
from ....._tools import create_repr, try_copy_to_list

if TYPE_CHECKING:
    from ....._types import OptStr, ExtendedParams
    from ..._enums._risk_type import RiskType
    from ..._enums._asset_class import AssetClass


class Definition(ContentProviderLayer):
    """

    Parameters
    ----------
    index_name : str, optional
        Example:
            "EURIBOR"
    main_constituent_asset_class : AssetClass, optional
        See detail class AssetClass.
    risk_type : RiskType, optional
        See detail RiskType class.
    currency : str, optional
        The currency code of the interest rate curve.
    curve_tag : str, optional
        User defined string to identify the curve. It can be used to link output results
        to the curve definition. Only alphabetic, numeric and '- _.#=@' characters
        are supported.
    id : str, optional
        Id of the curve definition
    name : str, optional
        The name of the interest rate curve.
    source : str, optional
        Example:
            "Refinitiv"
    valuation_date: str, optional
        Example:
            "2019-08-21"
    extended_params : dict, optional
        If necessary other parameters.

    Methods
    -------
    get_data(session=session, on_response=on_response, **kwargs)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, **kwargs)
        Returns a response asynchronously to the data platform

    Examples
    --------
    >>> from refinitiv.data.content.ipa.curves import zc_curve_definitions
    >>> definition = zc_curve_definitions.Definition(source="Refinitiv")
    >>> response = definition.get_data()

    Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        index_name: "OptStr" = None,
        main_constituent_asset_class: Optional["AssetClass"] = None,
        risk_type: Optional["RiskType"] = None,
        currency: "OptStr" = None,
        curve_tag: "OptStr" = None,
        id: "OptStr" = None,
        name: "OptStr" = None,
        source: "OptStr" = None,
        valuation_date: "OptStr" = None,
        extended_params: "ExtendedParams" = None,
    ) -> None:
        request_item = InterestRateCurveGetDefinition(
            index_name=index_name,
            main_constituent_asset_class=main_constituent_asset_class,
            risk_type=risk_type,
            currency=currency,
            curve_tag=curve_tag,
            id=id,
            name=name,
            source=source,
            valuation_date=valuation_date,
        )
        super().__init__(
            content_type=ContentType.ZC_CURVE_DEFINITIONS,
            universe=request_item,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)


DefnDefns = Union[Definition, List[Definition]]


class Definitions(ContentProviderLayer):
    """

    Parameters
    ----------
    universe : list of zc_curve_definitions.Definition
        See detail class zc_curve_definitions.Definition.

    Methods
    -------
    get_data(session=session, on_response=on_response, **kwargs)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, **kwargs)
        Returns a response asynchronously to the data platform

    Examples
    --------
    >>> from refinitiv.data.content.ipa.curves import zc_curve_definitions
    >>> definition1 = zc_curve_definitions.Definition(source="Refinitiv")
    >>> definition2 = zc_curve_definitions.Definition(source="Peugeot")
    >>> definitions = zc_curve_definitions.Definitions(
    ...     universe=[definition1, definition2]
    ...)
    >>> response = definitions.get_data()

    Using get_data_async
     >>> import asyncio
     >>> task = definitions.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        universe: "DefnDefns",
    ):
        universe = try_copy_to_list(universe)
        if not iterable(universe):
            universe = [universe]

        super().__init__(
            content_type=ContentType.ZC_CURVE_DEFINITIONS,
            universe=universe,
            __plural__=True,
        )

    def __repr__(self):
        return create_repr(self, class_name=self.__class__.__name__)



import json

import allure
import pytest

from refinitiv.data.content.ipa.curves import zc_curves
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.curves.conftest import (
    zc_curves_01,
    zc_curves_02,
    check_curve,
    check_curve_response,
    on_response,
)
from tests.integration.content.ipa.curves.zc_curve.conftest import zc_curves_universe
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - ZcCurve")
@allure.feature("Content object - ZcCurve")
@allure.severity(allure.severity_level.CRITICAL)
class TestZcCurve:
    @allure.title("Create a zc curve definition object with valid params and get data")
    @pytest.mark.caseid("37962810")
    @pytest.mark.smoke
    def test_create_zc_curve_definition_with_valid_params(self, open_session):
        response = zc_curves_01().get_data(
            on_response=lambda response, definition, session: on_response(
                response, definition, open_session
            )
        )

        check_curve_response(response, expected_currency="CHF")
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create a empty zc curve definition object and get data")
    @pytest.mark.caseid("37962856")
    def test_create_empty_zc_curve_definition(self, open_session):
        with pytest.raises(
            RDError,
            match="Error code QPS-Curves.6 | Invalid input: curveDefinition is missing",
        ):
            zc_curves.Definition().get_data()

    @allure.title("Create a zc curve definition object and check curve parameters")
    @pytest.mark.caseid("37962956")
    def test_create_zc_curve_definition_and_check_curve_parameters(self, open_session):
        response = zc_curves_01().get_data()
        curve = response.data.curve
        check_curve(curve)

    @allure.title("Create a zc curve definition object async")
    @pytest.mark.caseid("37963115")
    async def test_create_zc_curve_definition_async(self, open_session_async):
        valid_response, invalid_response = await get_async_response_from_definitions(
            zc_curves_02(), zc_curves.Definition()
        )
        curve = valid_response.data.curve

        check_curve(curve)
        check_curve_response(valid_response, expected_currency="EUR")

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="QPS-Curves.6",
            expected_error_message="Invalid input: curveDefinition is missing",
        )

    @allure.title("Create a zc_curve with extended parameters")
    @pytest.mark.caseid("C40202633")
    def test_get_zc_curves_with_extended_params(self, open_session):
        response = zc_curves.Definition(
            curve_definition=zc_curves.ZcCurveDefinitions(
                discounting_tenor="OIS",
            ),
            extended_params={
                "curveDefinition": {
                    "currency": "CHF",
                    "discountingTenor": "OIS",
                    "name": "CHF LIBOR Swap ZC Curve",
                },
                "curveParameters": {"useSteps": True, "valuationDate": "2019-08-21"},
            },
        ).get_data()

        check_curve_response(response, expected_currency="CHF")

        request = json.loads(response.request_message.content.decode("utf-8"))
        request_universe = request.get("universe")[0]
        assert (
            request_universe == zc_curves_universe
        ), f"Extended params are applied incorrectly"



import numpy as np

from .._content_data_validator import ContentDataValidator
from .._request_factory import DatesAndCalendarsResponseFactory
from ...._content_data import Data
from ...._content_data_provider import ContentDataProvider
from .....content.ipa._content_provider import DateScheduleRequestFactory
from .....delivery._data._data_provider import ValidatorContainer


class DateSchedule(Data):
    _dates = None

    @property
    def dates(self):
        if self._dates is None:
            self._dates = [np.datetime64(date) for date in self.raw["dates"]]

        return self._dates


date_schedule_data_provider = ContentDataProvider(
    request=DateScheduleRequestFactory(),
    response=DatesAndCalendarsResponseFactory(data_class=DateSchedule),
    validator=ValidatorContainer(content_validator=ContentDataValidator()),
)
